#! /usr/bin/env python
#===============================================================================
# views.viewplot
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

""" This module essentially define the :class:`ViewPlot` class to plot the data

"""
#===============================================================================
# python import
#===============================================================================
import wx
import os

#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import (HasTraits, Instance, Int, List, Bool, Str, Tuple, Either,
                        Button, Enum, Array, on_trait_change,
                        Property, Float, DelegatesTo)
from traitsui.api import (View, Item, HGroup)
from traitsui.menu import Action, ToolBar, Separator
from pyface.image_resource import ImageResource
from pyface.api import FileDialog, OK

#===============================================================================
# matplotlib, pylab, numpy
#===============================================================================
from matplotlib.figure import Figure, Axes
from matplotlib.colors import ColorConverter
import pylab as pl
import numpy as np

#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()

#===============================================================================
# masai imports
#===============================================================================
from masai.api import clipping, findregion
from .plot_editor import PlotEditor
from . import mpl as mpl

#===============================================================================
# constants and utilities
#===============================================================================
IMAGE_LOCATION = os.path.join(os.path.dirname(__file__), "..", "images")

#: conversion between cm and inches
cm2inch = 0.3937008

#: golden number
nb_or = 1.61803399

def mplcolor(color):
    """ Convert a wx color spec. to a matplotlib color spec. 
    
    Parameters
    -----------
    color : :class:`wx.color`
        A wx color definition
    
    Returns
    -------
    ncolor : tuple
        A matplotlib color
        
    
    """
    _color = np.array((1.0, 1.0, 1.0, 1.0))
    _color[:3] = np.asarray(color.Get()) / 255.
    return tuple(_color)

#: Available list of line style
lstyle = {'Solid': '-',
      'Dashed': '--',
      'Dashdot': '-.',
      'Dotted': ':',
      'None': ''}

#===============================================================================
# ViewPlot
#===============================================================================
class ViewPlot(HasTraits):
    """This class handles the plot feature and its view in the GUI application
    
    """
    fd = Instance('masai.fileio.source.Source')

    #: figure instance
    figure = Instance(Figure)
    
    #: size (w,h) of the figure
    #: default: 10 cm, 10cm/nb_or
    figsize = Tuple((10. * cm2inch, 10. * cm2inch / nb_or))
    
    #: dpi (dot per inch) for the figure
    #:     default: 96
    figdpi = Int(96)
    figbackground_color = Tuple(0.95, 0.95, 0.95)  
    # background color of the figure( matplotlib rgb color format)
    figlinewidth = Float(0.1)
    canvas = Property(depends_on='figure')

    # axes and scales
    ax = Instance(Axes)
    axs1 = Instance(Axes)
    axs2 = Instance(Axes)
    axf1 = Instance(Axes)
    axf2 = Instance(Axes)
    tscale = Enum('ms', 's', 'points')
    fscale = Enum('ppm', 'hz', 'points')
    xscale = Str('')
    yscale = Str('')
    xlim = Tuple
    ylim = Tuple
    fdmin = Float
    fdmax = Float

    #data
    x = Array(allows_none=True)
    y = Array(allows_none=True)

    # flags
    showproj = Bool(True)
    showsect = Bool(False)
    showpos = Bool(False)
    showcursor = Bool(False)

    show_model = DelegatesTo('fd')
    show_diff = DelegatesTo('fd')
    unmodified = DelegatesTo('fd')
    subtract = DelegatesTo('fd')
    
    # clipping
    is_clip_applied = Bool(True)
    applyclip = Button("Clip?")
    deleteclip = Button("Undo clip selection?")

    # zone fit
    showzone = Bool(False)

    # action states
    cursor_state = Bool(False)
    crosshairs_state = Bool(False)
    info_state = Bool(False)
    zoom_state = Bool(False)
    zone_state = Bool(False)
    clip_state = Bool(False)
    axzone = List
    axclip = List

    # Toolbar
    toolbar = Instance(ToolBar)

    # Toolbar actions.
    actions = List(Either(Action, Separator))

    # panel printer
    printer = Instance(mpl.FigurePrinter)
    POSTSCRIPT_PRINTING_COMMAND = 'lpr'

    location = Tuple()

    def default_traits_view(self):
        """Display the data and model plots.
        
        """ 
        return View(HGroup(
                           Item("applyclip", show_label=False,),
                           Item("deleteclip", show_label=False),
                           visible_when='not is_clip_applied',
                    ),
                    Item('figure',
                     editor=PlotEditor(),
                     show_label=False,
                     ),
                resizable=True ,
                toolbar=self.toolbar,
                width=600,
                height=500,
                )

    #===========================================================================
    # event router
    #===========================================================================
    @on_trait_change('fd.+')
    def _eventrouter(self, objet=None, name=None, old=None, new=None):
        """Is fired every time a trait associated to the source fd changes.
        wx.CallAfter is used to avoid multiple unecessary calls 
        
        """
        if self.fd.data_need_replot and self.fd.can_plot :
            logger.debug("...viewcontroller: update data plot")
            wx.CallAfter(self._plot1Ddata)
            self.fd.data_need_replot = False
            self._draw()

        if self.fd.model_need_replot and self.fd.can_plot :
            logger.debug("...viewcontroller: update modeldata plot")
            wx.CallAfter(self._plot1Dmodeldata)
            self.fd.model_need_replot = False
            self._draw()

    #===========================================================================
    # figure
    #===========================================================================
    def _figure_default(self):
        return Figure(figsize=self.figsize,
                      dpi=self.figdpi,
                      facecolor=self.figbackground_color,
                      edgecolor=mplcolor(wx.BLACK),
                      linewidth=self.figlinewidth
                      )

    def _get_canvas(self):
        fig = self.figure
        canvas = fig.canvas
        return canvas

    #===========================================================================
    # plot or update plot
    #===========================================================================
    def _resetplot(self):
        logger.debug('viewcontroller: _resetplot')
        n = self.fd.ndims
        if n == 1:
            self._reset1Dplot()
        elif n == 2:
            self._reset2Dplot()
        else:
            raise NotImplementedError, "Not yet: Cannot handle %dD "+\
            " plot higher than 2D" % n

    def _reset1Dplot(self):
        self._create1Daxes() 
        self._create1Dscales()
        self._plot1Ddata()

    def _reset2Dplot(self):
        self._create2Daxes()
        uc = self.fd.units
        self._create2Dscales(uc)
        self._plot2Ddata()

    def _clf(self):
        """Clear the figure
        
        """
        logger.debug('clear figure')
        self.figure.clf()
        self.x = []
        self.y = []

    def _create1Daxes(self):
        """Creates the basic 1D axes of the plot
        
        """
        logger.debug('viewcontroller: _create1Daxes')
        self._clf()
        self.ax = ax = self.figure.add_subplot(111)
        logger.debug('...1D axes created %s' % ax)

    def _create2Daxes(self):
        """Creates the basic axes of the plot
        
        """
        self._clf()
        if not self.showproj: 
            self.ax = ax = self.figure.add_subplot(111)

        else:
            left, width = 0.1, 0.65
            bottom, height = 0.1, 0.65
            bottom_h = left_h = left + width + 0.02
            rectct = [left, bottom, width, height]
            rectf2 = [left, bottom_h, width, 0.2]
            rectf1 = [left_h, bottom, 0.2, height]
            self.ax = ax = self.figure.add_axes(rectct)
            self.axf2 = axf2 = self.figure.add_axes(rectf2, sharex=ax)
            self.axf1 = axf1 = self.figure.add_axes(rectf1, sharey=ax)
            self.axs2 = axs2 = self.figure.add_axes(rectct, sharex=ax)
            self.axs1 = axs1 = self.figure.add_axes(rectct, sharey=ax)

            # no labels
            pl.setp(axf2.get_xticklabels(), visible=False)
            pl.setp(axf2.get_yticklabels(), visible=False)
            pl.setp(axf1.get_xticklabels(), visible=False)
            pl.setp(axf1.get_yticklabels(), visible=False)
            pl.setp(axs2.get_xticklabels(), visible=False)
            pl.setp(axs2.get_yticklabels(), visible=False)
            pl.setp(axs1.get_xticklabels(), visible=False)
            pl.setp(axs1.get_yticklabels(), visible=False)

        logger.debug('2D axes created %s' % ax)

    def _create1Dscales(self):
        """Create new scales

        """
        logger.debug('viewcontroller: _create1Dscales')
        if not hasattr(self,'ax') or self.ax is None:
            logger.debug("ax does not exist yet")
            self._resetplot()

        if self.fd.is_freq[-1]: 
            xscale = self.fscale
        else: 
            xscale = self.tscale

        uc = self.fd.units  
        self.x = getattr(uc[-1], "%s_scale" % xscale)() 
        self.xlim = xlim = getattr(uc[-1], "%s_limits" % xscale)()
        self.ax.set_xlabel(xscale)
        if self.fd.is_freq[-1]: 
            self.ax.set_xlim(xlim[::-1])    
                            # reverse the spectrum plot if it is transformed
                            # as to follow nmr convention
        else:
            self.ax.set_xlim(xlim)

    def _create2Dscales(self, uc):
        """Create new 2D scales

        """
        self._create1Dscales(uc) # for the F2 dimension

        if self.fd.is_freq[-2]: 
            yscale = self.fscale
        else: 
            yscale = self.tscale

        self.y = getattr(uc[-2], "%s_scale" % yscale)()
        self.ylim = getattr(uc[-2], "%s_limits" % yscale)()
        self.ax.set_ylabel(yscale)
        if self.fd.is_freq[0]: 
            self.ax.set_ylim(self.ylim[::-1]) 
            # reverse the spectrum if it is transformed
        else:
            self.ax.set_ylim(self.ylim)  

    def _get_line_from_label(self, label):
        """Knowing the label of line, return the line2D object

        """
        if self.ax is None:
            return False

        for line in self.ax.lines:
            if line.get_label() == label:
                return line
        return False

    def updatelineattr(self, label='experimental', ydata = None):
        """Set the attibute of a given line adressed through its label

        """
        # retrieve the line2D from its name
        line2d = self._get_line_from_label(label)

        if not line2d: 
            # if the line doesn't exist, we have nothing to do here
            return

        #check if the line attributes are in the directory
        #else, we should add them 
        if self.fd.line_attributes.has_key(label):
            line2d.set_linestyle(lstyle[self.fd.line_attributes[label][0]])
            line2d.set_linewidth(self.fd.line_attributes[label][1])
            line2d.set_color(mplcolor(self.fd.line_attributes[label][2]))
            if self.fd.subtract and label == 'baseline':
                # make it disappear when it is subtracted from the data
                line2d.set_linestyle('None')

        else:
            color = line2d.get_color()
            c = []
            for item in ColorConverter().to_rgba(color):
                c.append(int(np.round(item*255)))
            self.fd.line_attributes[label]=[
                        'Solid',0.65, wx.Colour(c[0],c[1],c[2],c[3]), -25, 0]

        if ydata is None:
            # get the line data (which may already contain an offset
            # in principle contained in the last attribute [4]
            prevoffset =self.fd.line_attributes[label][4]
            ydata = line2d.get_ydata()-prevoffset

        # offset given in % of the experimental data
        offpc =self.fd.line_attributes[label][3]
        y = self.fd.data.real
        offset = (np.max(y)-np.min(y))*offpc/100.

        # update the ydata according to the offset
        line2d.set_ydata(ydata + offset)

        #update the previous offset
        self.fd.line_attributes[label][4] = offset

        # visibility
        if label != 'experimental':
            line2d.set_visible(self.show_model)
        if label == 'difference':
            line2d.set_visible(self.show_diff)

        self._draw()

    def _plot1Ddata(self):
        """Plot or update experimental data
        .. todo:: 
            gives the possibility to have several experimental plots

        """
        # get the data to plot
        x = self.x
        y = self.fd.data.real.copy()

        # performs some checkin gon the validity of these data 
        #(x and y should be of same size)
        if x.size != y.shape[-1]:
            self._resetplot()
            return

        # get the line if it already exist or plot it
        line2d = self._get_line_from_label('experimental')
        if not line2d:
            # as the line does not exist we create it for the fist time
            line2d, = self.ax.plot(x, y, 'k', label='experimental')

        # now update the attributes and set the data according to the offset
        if self.subtract:
            y = y - self.fd.modeldata.real[0]
        self.updatelineattr('experimental', ydata = y)

        line2d = self._get_line_from_label('original')
        if not line2d and self.unmodified:
            # as the line does not exist we create it for the fist time
            line2d, = self.ax.plot(self.fd.xorig, self.fd.data_orig.real, 'r', label='original',zorder=-1)
            line2d.set_linestyle(':')
        elif line2d and not self.unmodified:
            line2d.remove()

    def _plot1Dmodeldata(self):
        """Plot or update model data  + baseline

        """
        if not self.fd.fittable:
            return
        
        # get the model data to plot
        x = self.x
        y = self.fd.modeldata.real.copy() 

        # check the validity of the data
        if y.size == 0:
            #no modeldata to plot yet
            return

        if x.size != y.shape[-1]:
            # replot the data first
            # indeed, we need to rescale as x !=y')
            self._plot1Ddata()
            x = self.x

        if not hasattr(self, 'ax'):
            #no ax ... return as nothing can be definitely done (no data)
            return

        # get the previously drawn lines 
        # (include also experimental and difference line)
        modellines = self.ax.get_lines()
        logger.debug("current plotted lines %s"%str(
                                    [line.get_label() for line in modellines]))

        # get the list of name of the model data to plot
        names = self.fd.modelnames
        logger.debug('list of lines to plot %s'%str(names))

        # check if we need to add or removes some line or just update
        for line in modellines:
            # check removing first
            label = line.get_label()
            if label not in ['experimental', 'baseline', 
                             'difference', 'modelsum'] and  label not in names:
                # this line should be removed
                line2d = self._get_line_from_label(label)
                line2d.remove()
                # recursive call to make a full cleaning of the plot
                self._plot1Dmodeldata()
                return

        nm = y.shape[0]
        if nm > 0:
            if self.fd.forceupdatemodel:
                # the model is not yet updated, we nned to do this before plotting
                return
            
            # model lines need to be updated or created
            for i in range(nm):
                line2d = self._get_line_from_label(names[i])
                if not line2d:
                    # as the line does not exist we create it for the fist time
                    try:
                        line2d, = self.ax.plot(x, y[i], ':', label=names[i])
                    except ValueError:
                        return
                    
                # now update the attributes 
                # and set the data according to the offset
                if names[i]=='modelsum' and self.fd.subtract:
                    y[i] = y[i] - self.fd.modeldata.real[0]

                self.updatelineattr(names[i], ydata = y[i])

            # display the difference
            diff = self.fd.data.real - self.fd.modeldata.real[-1]
            line2d = self._get_line_from_label("difference")
            if not line2d:
                # we plot this new line
                line2d, = self.ax.plot(x, diff, label="difference")
            self.updatelineattr('difference', ydata = diff)

    def _plot2Ddata(self):
        """
        
        """        
        x = self.x
        y = self.y
        z = self.fd.data.real
        self.ax.contour(x, y, z)
        if self.showproj:
            f2 = np.sum(z, axis=0)
            f1 = np.sum(z, axis= -1)
            self.axf2.plot(x, f2)
            self.axf1.plot(f1, y)
        # plot model lines
        if self.show_model:
            raise Exception(' DRAW 2D MODEL : not yet implemented')

    #===========================================================================
    # _printer_default
    #===========================================================================

    def _printer_default(self):
        """ Set the printer definition
        
        """
        pData = wx.PrintData()
        pData.SetPaperId(wx.PAPER_A4)
        if callable(getattr(pData, 'SetPrinterCommand', None)):
            pData.SetPrinterCommand(self.POSTSCRIPT_PRINTING_COMMAND)
        return mpl.FigurePrinter(self.canvas, pData)

        #TODO: add menu for printing

    #===========================================================================
    # draw
    #===========================================================================
    def _draw(self, resize=None):
        """Draw the associated figure onto the screen.
        Performs autoscaling if required

        """
        if self.ax is None: 
            logger.debug("...cannot draw because ax is none")
            return

        if (self.fd.autoscale or\
                    (resize == 'vert')) and \
                    (resize != 'hor') and\
                    self.fd.ndims == 1:  

            # Get the maximum and minimum of all displayed lines
            # calculation is done in the fitting zone
            # which should be the whole spectral width if there is no selection 
            uc = self.fd.units
            x = uc[-1].ppm_scale()
            fitzone = findregion(x, self.fd.zone)

            lines = self.ax.get_lines()
            ma = 0.0
            mi = 0.0
            for line2d in lines:
                y = line2d.get_ydata().take(fitzone)
                ma = max(ma,np.max(y))
                mi = min(mi,np.min(y))

            # we allow 2.5% below and above the lines
            am = (ma - mi) * 0.025
            self.ax.set_ylim(mi - am, ma + am)

        if (resize == 'hor'):
            if self.fd.is_freq[-1]: 
                self.ax.set_xlim(self.xlim[::-1])    
                # reverse the spectrum plot if it is transformed

        if (resize == 'vert' and self.fd.ndims == 2):
            if self.fd.is_freq[-2]: 
                self.ax.set_ylim(self.ylim[::-1])    
                # reverse the spectrum plot if it is transformed

        if self.canvas is not None:
            wx.CallAfter(self.canvas.draw)

    def _toolbar_default(self):
        toolbar = ToolBar(*self.actions)
        toolbar.image_size = (24, 24)
        toolbar.show_tool_names = False
        toolbar.show_divider = True
        return toolbar     

    #===========================================================================
    # _actions_default
    #===========================================================================
    def _actions_default(self):

        zoom = \
            Action(
                image=ImageResource('zoom.png', search_path=[IMAGE_LOCATION, ]),
                tooltip="Enable/disable zoom mode",
                style='toggle',
                checked_when='zoom_state',
                defined_when='True',
                enabled_when='not zone_state and not clip_state',
                perform=self._perform_zoom,
            )

        vresize = \
            Action(
                image=ImageResource('varrowsize.png', 
                                    search_path=[IMAGE_LOCATION, ]),
                tooltip="Vertical autoscale",
                defined_when='True',
                enabled_when='not zoom_state and not clip_state'+\
                                '  and not zone_state',
                perform=self._perform_vresize,
            )

        hresize = \
            Action(
                image=ImageResource('harrowsize.png', 
                                    search_path=[IMAGE_LOCATION, ]),
                tooltip="Horizontal autoscale",
                defined_when='True',
                enabled_when='not zoom_state and not clip_state'+\
                                ' and not zone_state',
                perform=self._perform_hresize,
            )         

        zone = \
            Action(
                image=ImageResource('zone.png', search_path=[IMAGE_LOCATION, ]),
                tooltip="Select a fitting region",
                defined_when='True',
                enabled_when='not zoom_state and not clip_state',
                style='toggle',
                checked_when='zone_state',
                perform=self._perform_zone,
            )

        clip = \
            Action(
                image=ImageResource('cutter.png', 
                                    search_path=[IMAGE_LOCATION,]),
                tooltip="horizontally clip spectra to a selected region",
                defined_when='True',
                enabled_when='not zoom_state and not zone_state',
                style='toggle',
                checked_when='clip_state',
                perform=self._perform_clip,
            )  

        info = \
            Action(
                image=ImageResource('info.png', search_path=[IMAGE_LOCATION, ]),
                tooltip="Show/hide info reporting",
                style='toggle',
                checked=False,
                defined_when='True',
                enabled_when='True',
                perform=self._perform_info,
            )

        do_print = \
            Action(
                image=ImageResource('printer.png',
                                    search_path=[IMAGE_LOCATION, ]),
                tooltip="Print figure",
                defined_when='True',
                enabled_when='True',
                perform=self._perform_do_print,
            )

        return [Separator(), zoom,
                             hresize,
                             vresize,
                Separator(), clip,
                Separator(), zone,
                Separator(), info,
                Separator(), do_print]

    def _perform_zoom(self):
        self.zoom_state = not self.zoom_state
        self.canvas.set_zoom(self.zoom_state)
        self.canvas.set_crosshairs(self.zoom_state)
        self.canvas.set_selection(self.zoom_state)

    def _perform_vresize(self):
        self._draw(resize='vert')

    def _perform_hresize(self):
        self._draw(resize='hor')

    def _perform_autoscale(self):
        self.autoscale = True
        self._draw()

    def _perform_info(self):
        self.info_state = not self.info_state
        self.canvas.set_location(self.info_state)

    def _perform_do_print(self):
        #self.printer.printFigure(self.figure)
        dlg = FileDialog(parent=self.canvas, action='save as', wildcard="*.pdf")
        if dlg.open() == OK:
            self.figure.savefig(dlg.path, dpi=300, 
                            orientation='paysage', papertype='A4', format='pdf')
#        savefig(fname, dpi=None, facecolor='w', edgecolor='w',
#        orientation='portrait', papertype=None, format=None,
#        transparent=False, bbox_inches=None, pad_inches=0.1)

    def _perform_zone(self):
        self.zone_state = not self.zone_state
        self.clip_state = False
        self.canvas.set_zone(True) 
        mpl.EVT_SELECTION(self.figure.canvas, -1, self._onselectfitzone)

    def _perform_clip(self): 
        self.clip_state = not self.clip_state
        self.zone_state = False
        self.canvas.set_zone(self.clip_state) # same procedure as for zone
        mpl.EVT_SELECTION(self.figure.canvas, -1, self._onselectclipzone)

    def _onselectfitzone(self, val):
        limits = self.ax.get_xlim()
        xmin = val.x1data
        xmax = val.x2data

        if xmin > xmax:
            xmin, xmax = (xmin, xmax)[::-1]

        x = self.x
        indmin, indmax = np.searchsorted(x, (xmin, xmax))
        indmax = min(len(x) - 1, indmax)

        self.fd.zone = zone = list((x[indmin], x[indmax]))
        
        for item in self.axzone:
            item.set_visible(False)

        # warning the spectrum is reversed, this is why we invert the limits
        self.axzone.append(self.ax.axvspan(limits[-1], zone[0], 
                                facecolor='lightyellow', hatch='/', alpha=0.3))
        self.axzone.append(self.ax.axvspan(zone[-1], limits[0], 
                                facecolor='lightyellow', hatch='\\', alpha=0.3))
                
        wx.CallAfter(self._draw)

        self.zone_state = False
        self.canvas.set_zone(False)     
        mpl.EVT_SELECTION(self.figure.canvas, -1, None)

        self.fd.forceupdatemodel = True

    def _onselectclipzone(self, val=None, delete=False):

        limits = self.ax.get_xlim()
        if not delete:
            xmin = val.x1data
            xmax = val.x2data
        else:
            xmin = min(limits)
            xmax = max(limits)

        if xmin > xmax:
            xmin, xmax = (xmin, xmax)[::-1]

        x = self.x
        indmin, indmax = np.searchsorted(x, (xmin, xmax))
        indmax = min(len(x) - 1, indmax)

        self.fd.clip = clip = list((x[indmin], x[indmax]))
        for item in self.axclip:
            item.set_visible(False)

        # warning the spectrum is reversed, this is why we invert the limits
        self.axclip.append(self.ax.axvspan(limits[-1], clip[0], facecolor='red', hatch='.', alpha=0.8))
        self.axclip.append(self.ax.axvspan(clip[-1], limits[0], facecolor='red', hatch='.', alpha=0.8))
        wx.CallAfter(self._draw)

        self.clip_state = False
        self.canvas.set_zone(False)     
        mpl.EVT_SELECTION(self.figure.canvas, -1, None)

        self.is_clip_applied = False
        
    def _applyclip_fired(self):
        self.fd.can_plot = False
        clipping(self.fd)
        uc = self.fd.units
        self.fd.clip = list(uc[-1].ppm_limits())
        #FIXME: check this because when apply several times, low limits changes!!!!
        self.fd.sourcepref.set('processing.clip', self.fd.clip)
        self.fd.sourcepref.flush()
        self.fd.can_plot = True
        self.is_clip_applied = True
        wx.CallAfter(self._draw)
                
    def _deleteclip_fired(self):
        self._onselectclipzone(delete=True)
        self.fd.sourcepref.set('processing.clip', self.fd.clip)
        self.fd.sourcepref.flush()
        self.is_clip_applied = True

#we could also use
#self.printer.pageSetup()
#self.printer.previewFigure(self.get_figure())    



#! /usr/bin/env python
#===============================================================================
# modelling.sequence
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

""" docstring """
import os

from traits.api import (HasTraits, Constant,
                        Instance, Enum, List, Code, Button, Str)
from traitsui.api import (View, Item, TextEditor, CodeEditor, HGroup,
                        VGroup, Group, spring, Label, EnumEditor)
from traitsui.menu import OKButton, CancelButton

_scriptdir =os.path.join(os.path.dirname(__file__),'script')
_pulse_sequence_list= os.listdir(_scriptdir)

class SequenceAvailable(HasTraits):
    type = List(_pulse_sequence_list)
sequence_available= SequenceAvailable()

class PulseSequence(HasTraits):
    
    type = Str('ideal_pulse')
    available = Constant(sequence_available)
    
    simpson_script = Code()
    script = Code()
    
    new = Button('NEW')
    dup = Button('DUPLICATE')
    rem = Button('DELETE')
    
    def _script_default(self):
        with open(os.path.join(_scriptdir, self.type), 'rb') as f:
            return f.read()
            
    def _type_changed(self):
        with open(os.path.join(_scriptdir, self.type), 'rb') as f:
            self.script = f.read()
            
    _name = Str()
    
    def _dup_fired(self):
        _dup_view = View( Item('_name', editor = TextEditor(enter_set=True),
                               label= 'name of the duplicate',),
                         buttons = [OKButton, CancelButton], width=.33)
        self._name='copy of %s'%self.type
        done = self.edit_traits(view=_dup_view, kind='modal')
        if done.result: # not cancel
            self._name = self._name.replace(' ','_')
            with open(os.path.join(_scriptdir, self._name), 'wb') as f:
                f.write(self.script)
                self.available.type.append(self._name)
            self.type = self._name

    def _new_fired(self):
        _new_view = View( Item('_name', editor = TextEditor(enter_set=True),
                               label= 'name of the new sequence',),
                         buttons = [OKButton, CancelButton], width=.33)
        self._name='new name'
        done = self.edit_traits(view=_new_view, kind='modal')
        if done.result: # not cancel
            self._name = self._name.replace(' ','_')
            with open(os.path.join(_scriptdir, self._name), 'wb') as f:
                f.write(self.script)
                self.available.type.append(self._name)
            self.type = self._name    
    
    def _rem_fired(self):
        pass
        #TODO
         
    def default_traits_view(self):
        return View(
                    Item(
                        'type',
                        editor = EnumEditor(
                                   name = 'object.available.type'),
                        label = 'sequence script'
                        ),
                    VGroup( 
                          Item(
                               'script', 
                                show_label = False
                               ),
                           ),
                    HGroup(spring,
                           Item('new', show_label=False,),
                           Item('dup', show_label=False,),
                           Item('rem', show_label=False,),
                           ),
                resizable = True,
                )

        
#===============================================================================
if __name__ == '__main__':
    PulseSequence().configure_traits()
    
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#===============================================================================
# masaigui
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""

The :mod:`masaigui` module is the main entry point of the **MASAI** 
GUI (Graphical User Interface) application.

The GUI can be started in the terminal by typing ::

    $ python masaigui.py

assuming that the current directory is the **masai** application directory .

Some copyright information and the version number are witten in the terminal 
while the application is loading. 

To display some debugging messages while running **MASAI** one can use the 
*--debug* flag::

    $ python masaigui.py --debug
    
The cost of this may be a significant slowing down of the program.

"""

#===============================================================================
# print version number at the start of the program
#===============================================================================
from version import _TITLE_, _VERSION_, _REVISION_, _COPYRIGHT_
print "%s\nversion : %s-%s\nCopyrigth (c)%s" % (_TITLE_, _VERSION_, 
                                                _REVISION_, _COPYRIGHT_)

#===============================================================================
# python import
#===============================================================================
import os
import sys
import logging
import time

#===============================================================================
# logger
#===============================================================================
logger = logging.getLogger()
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)
for arg in sys.argv[1:]:
    if arg == '--debug': 
        logger.setLevel(logging.DEBUG)
        logger.debug('Masai is running in DEBUGGING mode.')

#===============================================================================
# PYTHONPATH
#===============================================================================
# in case masai was not yet installed using setup 
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))
logger.debug(sys.path)

#===============================================================================
# Enthought imports
#===============================================================================
from traits.api import (HasTraits, Instance, Property, DelegatesTo,
                        on_trait_change, Str, HTML)
from traitsui.api import (View, HSplit, Handler, auto_close_message,
                          ImageEditor, Group, Label, Item, UItem, HTMLEditor)
from traitsui.menu import MenuBar, Menu, Action
from pyface.api import ImageResource, SplashScreen
from pyface.api import confirm, YES
from pyface.api import GUI
from apptools.preferences.api import (set_default_preferences,
                                      get_default_preferences, Preferences)

#===============================================================================
# masai imports
#===============================================================================
from masai.api import Source, efp, largest_power_of_2
from masai.controllers import Controllers
from masai.views import ViewPlot
from masai.options import edit_options
from masai.utils import InstanceUItem
from masai.version import _COPYRIGHT_, _VERSION_,_REVISION_

#===============================================================================
# splash screen and GUI (#TODO: make a pyface application)
#===============================================================================
splash_screen = SplashScreen(image=ImageResource('images/splash'))
_gui = GUI(splash_screen=splash_screen)

#===============================================================================
# constants
#===============================================================================
license_label = \
"""
<html><body>
<H2>Masa&iuml;, a framework for processing and modelling of solid state NMR spectra</H2>

<b>Copyright: &copy; %(copyright)s</b>

<b>version: %(version)s.%(revision)s</b>

<b>email: </b> <a href='mailto:christian.fernandez@ensicaen.fr'>
christian.fernandez@ensicaen.fr</a>
 
This software is a computer program whose purpose is to perform 
processings and analysis of solid-state NMR spectra.
 
This software is governed by the CeCILL-B license under French law and abiding 
by the rules of distribution of free software.  
You can  use, modify and/or redistribute the software under the terms of 
the CeCILL-B license as circulated by CEA, CNRS and INRIA at the 
following URL: <a href='http://www.cecill.info'>http://www.cecill.info</a>. 
 
As a counterpart to the access to the source code and rights to copy, modify 
and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of 
the economic rights,  and the successive licensors  have only  limited
liability. 
 
In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 
 
The fact that you are presently reading this means that you have had
knowledge of the CeCILL-B license and that you accept its terms.
 
</body></html>
"""%{'copyright': _COPYRIGHT_, 'version':_VERSION_, 'revision':_REVISION_}
  
contact_label = """
http://github.com/fernandezc/masai
"""
pathini = os.path.join(os.path.expanduser('~'), '.masai')

default_ini = """
[application]
last_source_file = 
path_of_data = 
width = 1.0
height = 0.8

[processing]
process_automatically = True

[plot]
bgcolor = "(255, 255, 255, 255)"
fgcolor = "(0, 0, 0, 255)"
"""

#===============================================================================
# prepare ini
#===============================================================================
if not os.path.exists(pathini):
    logger.debug('create hidden directory %s' % pathini)
    os.mkdir(pathini)

inifile = os.path.join(pathini, 'masai.ini')
if not os.path.exists(inifile): 
    logger.info('open masai.ini in %s' % pathini)
    with open(inifile, 'w') as f:
        f.write(default_ini)  

set_default_preferences(Preferences(filename=inifile))

#===============================================================================
# MasaiGuiMainWindow
#===============================================================================
class MasaiGuiMainWindow(HasTraits):
    """ The Masai main GUI window.
    
    here go the instructions to create and destroy the application.
    General menu are also defined here.

    """
    
    #: Instance of :class:`ViewPlot`, where the plot of the experimental data 
    #: and modelled data are displayed 
    viewplot = Instance(ViewPlot)
    
    #: Instance of :class:`Controllers`, allowing action on the data and the plot view.
    controllers = Instance(Controllers)
    
    #: Instance of the :class:`Source`, which store the experimental data and the modelled data as well.
    fd = Instance(Source)
    
    #: Instance of :class:`Preferences`, which handle the main application preferences
    applicationpref = Instance(Preferences)

    #: Height of the main window (will be read from preference)
    height = Property
    
    #: Width of the main window (will be read from preference)
    width = Property
    
    # Are we loading the source and doing the first processings
    _loading = True
    
    # can we plot the data?
    can_plot = DelegatesTo('fd')

    # for the about dialog
    license = HTML(license_label)

    #===========================================================================
    # properties
    #===========================================================================
    def _get_height(self):
        """Height of the main window
        """
        return float(self.applicationpref.get('application.height'))

    def _get_width(self):
        """Width of the main window
        """
        return float(self.applicationpref.get('application.width'))

    #TODO: write the current value back to the preference.
    
    #===========================================================================
    # preferences
    #===========================================================================
    def _applicationpref_default(self):
        """Get the preferences for the application

        """
        return get_default_preferences()

    #===========================================================================
    # menus    
    #===========================================================================
    _open_action = Action(name="&New Source",
                          accelerator="Ctrl+N",
                          action="_performs_open_fd",
                          image=ImageResource("open.png"),
                          tooltip=u"Open a different dataset")
    
    _options_action = Action(name="&Options",
                          accelerator="Ctrl+O",
                          action="_performs_options",
                          image=ImageResource("options.png"),
                          tooltip=u"Open the preference manager")
    
    _exit_action = Action(name="&Quit",
                          accelerator="Ctrl+Q",
                          action="close",
                          image=ImageResource("exit")) 
    
    _help_action = Action(name="&Help",
                          accelerator="Ctrl+H",
                          action="_performs_show_help",
                          image=ImageResource("help"),
                          tooltip=u"Open help in an external web browser")

    _about_action = Action(name=u"About Masai",
                          action="_performs_show_about",
                          image=ImageResource("about"),
                          tooltip=u"Various information about Masa\u00EF")

    #: Menu entry used to show the About window or Help Browser.
    help_menu = Menu("|",
                     _help_action, "_",
                     _about_action,
                     name="&Help")

    #: Menu entry used to show the Preferences pages
    options_menu = Menu("|",
                     _options_action,
                     name="Options")

    #: Menu used to open new source dataset, print data or the close the application
    file_menu = Menu("|",
                     _open_action, '_',
                     _exit_action,
                     name="File")
    #: the main menubar
    menubar = MenuBar(file_menu, options_menu, help_menu)

    def _performs_open_fd(self):
        """Open a new source of data

        """
        fd = Source()
        fd = self._autoprocess(fd)

        #end of loading
        self._loading = False
        self.fd = fd

        # now we can plot
        self.can_plot = True

    def _performs_options(self):
        """Open the preference manager .

        """
        edit_options()

    def _performs_show_about(self):
        """ Display a view information: 'about Masai'

        """
        about_view = View(
                      Item( 'license',
                            show_label = False,
                            editor = ImageEditor(
                                    image = ImageResource( 'images/splash' ) ) 
                           ),
                      Group(UItem('license', 
                                 editor=HTMLEditor(format_text=True),
                                 width = 793,
                                 height = 300,
                                ), 
                            label="License"),
                      Group(Label(contact_label), label="Contribute"),
                    #    Group(Label(credits_label), label="Credits"),
                        title=u"About Masa\u00EF", buttons=["OK"],
                    #    icon=frame_icon
                    )
        self.edit_traits(kind="livemodal", view=about_view)

    def _performs_show_help(self):
        """Open the help pages

        """
        self._open_help()
        
    def _browser_open(self, url):
        """Open a Browser to display the help pages
    
        """
        if sys.platform == 'darwin':
                os.system('open %s &' % url)
        else:
            import webbrowser
            if webbrowser._iscommand('firefox'):
                firefox = webbrowser.get('firefox')
                firefox._invoke(['-chrome', url], remote=False, autoraise=True)
            else:
                webbrowser.open(url, autoraise=1)
    
    def _open_help(self):
        """ Open the masai user manual index in a browser.
    
        """
        url = "http://fernandezc.github.com/masai/doc/html/index.html"
        auto_close_message("Opening help in web browser...")
        self._browser_open(url)
        
    #===========================================================================
    # view
    #===========================================================================
    def default_traits_view(self):
        """Return the main view, with a left side containing the 
        controllers, and the right side the spectral view

        """
        return View(
                    HSplit(
                          InstanceUItem('controllers', width=.33,),
                          InstanceUItem('viewplot', width=.66,),
                          show_border=True,
                          visible_when='fd and fd.is_loaded and not _loading',),
                    title="%s (%s-%s)" % (_TITLE_, _VERSION_, _REVISION_) ,
                    resizable=True,
                    x=0,
                    y=20,
                    height=self.height,
                    width=self.width,
                    menubar=self.menubar,
                    handler=MasaiGuiWindowHandler(),
                   )

    #===========================================================================
    # _close
    #===========================================================================
    def _close(self):
        """Close the application

        """
        # TODO: performs some cleaning before closing
        print "exiting "
        sys.exit()

    #===========================================================================
    # events
    #===========================================================================
    @on_trait_change('fd')
    def _fd_updated(self):
        """The source fd has been changed. Pass the change to the panel
        and reset the ViewPlot
        """  

        if self.fd.is_loaded:
            logger.debug("main: source fd has changed")
            self.viewplot = ViewPlot(fd=self.fd)
            self.controllers = Controllers(fd=self.fd, viewplot=self.viewplot)

    #===========================================================================
    # __init__: initialisation
    #===========================================================================
    def __init__(self, **traits): 
        """
        Initialisation and setup of MasaiGuiMainWindow.

        The program try to read the last opened file 
        or present a dialog to select a new dataset  

        """
        super(MasaiGuiMainWindow, self).__init__(**traits)

        if self.applicationpref.get('application.last_source_file'):
            # open the last opened file if possible
            fd = Source('last')
        else:
            # ask which file to open
            fd = Source()

        if not fd.is_loaded:
            logger.error("Cannot load these data or the cancel button"+\
                         " was pushed.... program stopped")
            sys.exit()  
            #FIXME: avoid exiting the application for such stupid reason
        
        fd = self._autoprocess(fd)

        #end of loading
        self._loading = False
        self.fd = fd

        # now we can plot
        self.can_plot = True

    def _autoprocess(self, fd): 

        if eval(self.applicationpref.get('processing.process_automatically',
                                         'False')):
            # Fourier transform
            # but before we need to read the parameters 
            # from the source preference file 
            fd.TDeff[-1] = int(fd.sourcepref.get('processing.tdeff',
                                                 fd.TD[-1]))
            fd.SI[-1] = int(fd.sourcepref.get('processing.si',
                                            largest_power_of_2(fd.TD[-1]) * 2))
            fd.LB[-1] = float(fd.sourcepref.get('processing.lb', fd.LB[-1]))
            fd.PHC0[-1] = float(fd.sourcepref.get('processing.ph0',
                                                  fd.PHC0[-1])) % 360.0
            fd.PHC1[-1] = float(fd.sourcepref.get('processing.ph1',
                                                  fd.PHC1[-1]))
            fd.sourcepref.flush()

            efp(fd)

        return fd

#===============================================================================
# MasaiGuiWindowHandler
#===============================================================================
class MasaiGuiWindowHandler(Handler):
    """Handler for the guiwindow window: Essentially handle the exit process of the
    masai application

    """
    def close(self, info, _dummy=False):
        """Close the application after displaying a confirmation dialog
        
        """
        if info.initialized:
            retval = confirm(parent=info.ui.control,
                             message=u"Do you want to quit Masa\u00EF " + \
                             "(note: the current state is already saved)?",
                             title=u"Quit Masa\u00EF?",
                             default=YES)
            if retval == YES:
                info.object.controllers.fitcontroller.paramscript.timer.stop()
                info.object._close()
        return False

#===============================================================================
# run
#===============================================================================
def run():
    """Main entry to run the application
    
    """
    # TODO: Create a full pyface or envisage application
    # Create and open the main window.
    window = MasaiGuiMainWindow()

    time.sleep(5)
    _gui.start_event_loop() # A hack to close the splash screeen
    #TODO modify this ugly way to do it 
    #(use  a real pyface or envisage application)

    window.configure_traits()
   
#===============================================================================
if __name__ == "__main__":
    run()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#===============================================================================
# fileio.source
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================
"""
.. _source:

This module consist mainly in defining the class :class:`Source` which will contains
all data and parameters about the NMR experiments and the class 'ConvertUnits'.

It can be imported in a program as:

    >>> from masai.api import Source

"""

#===============================================================================
# Standard library imports.
#===============================================================================
import os
import string
import wx
from copy import deepcopy

#===============================================================================
# logger
#===============================================================================
import logging
logger = logging.getLogger()

#===============================================================================
# Enthought imports.
#===============================================================================
from traits.api import (HasTraits, Dict, Any, Directory, Bool, Property, Str,
                         Enum, Float, Array, on_trait_change, Instance, List,
                         cached_property)
from traitsui.api import (Item, View, Action, Handler, Group)
from apptools.preferences.api import (#set_default_preferences, 
                                      get_default_preferences, Preferences)

#===============================================================================
# Numpy 
#===============================================================================
import numpy as np
from numpy import nan

#===============================================================================
# Local imports
#===============================================================================
from masai.processing.processbase import _ft, _ift
from masai.fileio.nmrglue.bruker import read
from masai.database.isotopes import Isotopes

# module global definition
status = {0:[''], 1:['', '2'], 2:['', '2', '3']}  
# give X which is used to read the proper acquXs, procXs files 
# according the parmode

#===============================================================================
# _OpenHandler (private class)
#===============================================================================
class _OpenHandler(Handler):
    """ The handler for the open dialog
    (should not be called directly by end-users)
    """
    def close(self, info, is_OK):        
        info.object._is_ok = is_OK
        return True

#===============================================================================
# Base class of a source
#===============================================================================
class Source(HasTraits):
    """This class handles a Bruker NMR source of data

    It first reads Bruker binary (ser/fid) files, Bruker JCAMP-DX parameter 
    (acqus) files, and Bruker pulse program (pulseprogram) files. The object created 
    is the main 'model' (in the MVC concept) of the **Masaï** application. 

    Examples
    --------

    We will use a wild import here for masai and use pylab librairies

    >>> from masai.api import *
    >>> import pylab as pl

    To create a new instance of a Source

    >>> fd1 = Source(path = 'test/1') 

    Plotting the datas (using the Source.data attribute)       

    1D case:

    >>> fig1 = pl.figure(1)  
    >>> p = pl.plot(fd1.data.real) 

    or 2D case:

    >>> fd2= Source(path='test/2') 
    >>> fig2 = pl.figure(2)
    >>> c = pl.contour(fd2.data.real, 30) 
    >>> pl.xlim(0,1024)
    (0, 1024)
    >>> pl.ylim(0,100)
    (0, 100)

    To print nmr parameters extracted from brukers acqu's and proc's

    >>> sf1, sf2 = fd2.SFO1
    >>> sf2 # is the SFO1 in direct dimension
    104.31847052

    Do not forget to add this:

    >>> pl.show()

    to be able to see the matplotlib plots.

    See :ref:`apihowto` for more examples. 

    Parameters
    ----------
    
    path : string, optional
        Default is None, so that a dialog ui will be launched.
        if path='last', then the last opened NMR file will be opened.
        if path is defined as a valid (existing) Bruker NMR directory:
        i.e., such as '/user/nmr/name_of_experiment/expno' 
        then the file will be opened without any further dialog.

    Attributes
    ----------
    
    data : numpy ndarray
        Contains the ndarray representing the current state of the source data
    data_orig: numpy ndarray
        Contains the ndarray representing the source data 
        (before any transforms)

    Notes
    -----
    
    Preferences are updated, so that next call to this class should point
    on the last used NMR path.

    """    
    #===========================================================================
    # data and model
    #===========================================================================
    # The directory to read source from
    _directory = Directory('', allow_none=False)

    # check if the specified file is valid
    _is_valid = Property(Bool(False), depends_on='_directory')

    # the message to display when errors in the file open dialog
    _message = Property(Str, depends_on='_directory')
    
    # data input (as read by the __init__ function): Read only
    data_orig = Property(Array)
    _data_orig = Array
    xorig = Array
    
    # data output (can be modified): original data are stored in data_orig
    # attributes for the view plot are specified as named arguments
    data = Property(Array)
    _data = Array

    #ndims
    ndims = Property(depends_on='_data_orig')

    # is-it processing running?
    is_processing = Bool(False)
    
    # is_transposed (required for processing multidimensionnal spectra)
    is_transposed = Bool(False)

    # is_freq say if the spectrum is fourier transformed 
    # or not in the various axis
    is_freq = List(Bool)

    # This flag is used mainly by the clipping ptrocedure
    is_resized = Bool

    # model data (resulting from simulations)
    modeldata = Property(Array, allows_none=True)
    _modeldata = Array
    modelnames = List(Str)
    forceupdatemodel = Bool(False)
    A = Float 
    
    # units (handle of units for the axis)
    units = List

    # fit zone (zone taken into account for fitting)
    zone = List

    # can we fit
    fittable = Bool(False)
    
    # clip zone (clipping region)
    clip = List

    # Loading state of this source
    is_loaded = Bool(False)

    # is the replot required?
    data_need_replot = Bool(False)
    model_need_replot = Bool(False)
    
    # should we plot?
    can_plot = Bool(False)

    #===========================================================================
    # plot traits
    #===========================================================================
    # should we show original fid
    unmodified = Bool(False)
    
    # should we plot the model?
    show_model = Bool(True, 
                      desc='if we want to show the model lines in view plot')

    # should we plot the difference between model and experimental data
    show_diff = Bool(False,
                     desc='if we want to show difference between data '+\
                                                    'and model in view plot')
    # substract baseline
    subtract = Bool(False)
    
    # autoscaling plot
    autoscale = Bool(True, 
                     desc='an automatic vertical rescaling of the view plot')
    
    line_attributes = Dict(
            # style, width, color, offset %, previous %
            {'experimental':['Solid', 1.0, wx.Colour(0 , 0 , 0, 255), 0., 0.],
             'baseline':['Solid', 0.65, wx.Colour(255, 255 , 0, 255), 0., 0.],
             'modelsum':['Dashed',0.85, wx.Colour(255, 0 , 0, 255), 0., 0],
             'difference':['Dashdot',0.5, wx.Colour(0, 255 , 0, 255), -30.0, 0.],
             }
    )

    #===========================================================================
    # others
    #===========================================================================
    # Python traits
    _ = Any

    # list of of parameters attributes
    _attribs = List

    #===========================================================================
    # preferences
    #===========================================================================

    # main application preferences
    applicationpref = Instance(Preferences)

    #references for the current opened dataset
    sourcepref = Instance(Preferences)        

    def _applicationpref_default(self):
        """
        Get the preferences for the application

        """
        pref = get_default_preferences()
        if pref is None:
            # this will not be executed if the GUI is running
            # but is needed when one use the API only
            inifile = os.path.join(os.path.expanduser("~"),".masai","masai.ini")
            pref = Preferences(filename=inifile)
        return pref

    #===========================================================================
    # property handlers
    #===========================================================================
    
    @cached_property
    def _get__is_valid(self):
        if not os.path.exists(self._directory):
            return False
        ld = os.listdir(self._directory)
        item = 'fid'
        if item in ld:
            return True
        item = 'ser'
        if item in ld:
            return True        
        return False

    @cached_property
    def _get__message(self):
        if not os.path.exists(self._directory):
            return 'Please, choose a valid Bruker directory containing ' + \
                    'either a fid or a ser file...'
        elif self._is_ok:
            return "loading..."
        if not self._is_valid:
            return 'Sorry, but this directory ' + \
                   'does not contain a valid Bruker fid or ser file'
        else:
            return ''

    def _get_data_orig(self):
        """Read only property
        """
        return self._data_orig.copy()

    def _get_data(self):
        return self._data
    def _set_data(self, value):
        self._data = value

    @cached_property
    def _get_ndims(self):
        """Read only property
        """
        return len(self._data_orig.shape)

    def _set_modeldata(self, value):
        if value is None:
            self._modeldata = np.empty()
        else:
            self._modeldata = value
    def _get_modeldata(self):
        return self._modeldata

    #===========================================================================
    # default_traits_view
    #===========================================================================
    def default_traits_view(self):
        """
        Return the default view of the source traits

        """
        #TODO: could be improved for testing for instance
        return View(
            Item('_directory', style='readonly'),
            Group(
              Item('TD', style='readonly')
            ),
            title="Source of data",
            width=300,
            resizable=True,
        )

    #===========================================================================
    # _load_view
    #===========================================================================
    _is_ok = Bool(False)
    def _load_view(self):
        """ View to choose data location

        """
        return View(
            Item('_directory', style='simple'),
            Item('_message', style='readonly', show_label=False),
            buttons=[
                       Action(name='OK', enabled_when='_is_valid'),
                       'Cancel'
                       ],
            title="Locate Bruker's Sources or Ser's",
            width=600,
            resizable=True,
            handler=_OpenHandler(),
            )
    
    def reset(self):
        """Here the source is reset to its original
        
        """
        self._is_ok = True
        self._import_bruker_source(reset=True)
        
    #===========================================================================
    # _import_bruker_source
    #===========================================================================
    def _import_bruker_source(self, reset=False):
        """
        Import bruker data and parameters

        """
        self.is_loaded = False

        if (self._is_valid and self._is_ok):

            # read the data
            p, self._data_orig = read(self._directory)

            # normalisation of amplitude to ns=1K and rg=1
            self._data_orig = (self._data_orig / float(p['acqus']['NS']) 
                               / float(p['acqus']['RG']))

            # make this file the last in preferences file
            self.applicationpref.set('application.last_source_file',
                                     self._directory)

            # create a table to handle all parameters
            self._attribs = []

            # create attributes from parameters in the form self.PARAMETER
            self._make_parameters(p)

            #  when just read the data are obviously non transposed       
            self.is_transposed = False

            # and the data are not transformed, so is_freq must be false
            self.is_freq = [False for _i in self._data_orig.shape]  

            # put this attribute in the list of attribs 
            self._attribs.append('is_freq')

            #correct immediately the digital filter
            self._remove_digital_filter()

            # make the working ndarray from the original data
            self.data = self.data_orig  
            self.make_units()
            self.xorig = self.units[-1].ms_scale()
            
            # create some labels for the dimensions
            self._make_labels()

            # modify nuclei informations
            for i in range(self.ndims):
                nucl = self.NUC1[i]
                self.NUC1[i] = nuclei = Isotopes(nucl)   
                logger.info(
                   "This experiment involves %s (I=%s) in %s dimension" % (nucl,
                                                nuclei.spin, self.dimlabels[i]))

            # create the preferences for this source if necessary            
            self._create_source_pref(reset)

            # write preference changes for the application
            self.applicationpref.flush()

            # end the initialization
            logger.debug("Reading data done successfully")    
            self.is_loaded = True
            return True

    #===========================================================================
    # _make_parameters
    #===========================================================================
    def _make_parameters(self, p):
        """Create self attributes from the NMR parameter dictionary
        """
        parmode = p['acqus']['PARMODE']

        for keys, value in p.iteritems():
            setattr(self, keys, value)  
            
        # create list of parameters with the dimension size 
        # (direct dimension is the last one)
        # and store the attributes name in _attribs
        for key in getattr(self, 'acqus').iterkeys():
            # print key
            setattr(self, key, 
                    [getattr(self, 'acqu%ss' % status[parmode][i])[key]
                                for i in range(parmode, -1, -1)])
            self._attribs.append(key)

        for key in getattr(self, 'procs').iterkeys():
            setattr(self, key, 
                    [getattr(self, 'proc%ss' % status[parmode][i])[key] 
                                for i in range(parmode, -1, -1)])
            self._attribs.append(key)
            # create also a save attribute
            savkey = "%sSAV" % key
            setattr(self, savkey, [0.0 for i in range(parmode, -1, -1)])
            self._attribs.append(savkey)

    #===========================================================================
    # make_units
    #===========================================================================
    def make_units(self):
        """recalculate units
        """
        logger.debug("make units")
        self.units = []
        try:
            for i in range(self.ndims):
                self.units.append(ConvertUnits(self, i))
            if 'units' not in self._attribs: 
                self._attribs.append('units')
        except: 
            # this is more likely because data are not yet fully loaded
            pass
        
        return self.units

    #===========================================================================
    # _make_labels
    #===========================================================================
    def _make_labels(self):
        """create some labels for the dimensions
        """
        self.dimlabels = []
        for i in range(self.ndims):
            if i == self.ndims - 1: 
                self.dimlabels.append('direct')
            else:
                #TODO: implement 3D (in a further version)
                self.dimlabels.append('F1')    
        if not hasattr(self, 'dimlabels'): self._attribs.append('dimlabels')

    #===========================================================================
    # _create_source_pref
    #===========================================================================
    def _create_source_pref(self, reset=False):
        inifile = os.path.join(self._directory, 'preferences.ini')
        
        # erase inifile if reset is required
        if os.path.exists(inifile) and reset:
            os.remove(inifile)
            
        _new = False
        if not os.path.exists(inifile):
            # first time created
            _new = True
        self.sourcepref = Preferences(filename=inifile)
        if _new:
            self.sourcepref.set('processing.forcefid', True)
        self.sourcepref.flush()

    #===========================================================================
    # listing
    #===========================================================================
    def listing(self):
        """List NMR parameters for nice printing
        """

        strg = "--acquisition--\n"
        for key in getattr(self, 'acqus').iterkeys():
            strg += "%s: %s\n" % (key, getattr(self, key))

        strg += "\n--processing--\n"
        for key in getattr(self, 'procs').iterkeys():
            strg += "%s: %s\n" % (key, getattr(self, key))

        strg += "\n--pulse program--\n"
        for key in getattr(self, 'pprog').iterkeys():
            strg += "%s: %s\n" % (key, getattr(self, key))

        return strg

    #===========================================================================
    # detect changes
    #===========================================================================
    @on_trait_change('+,+[]')
    def _eventrouter(self, objet=None, name=None, old=None, new=None):
        
        if name.startswith('_data') or name=='_modeldata':
            
            # something has changed  with the source data
#            logger.debug("Source : %s from shape:%s to shape %s"%(name, 
#                                                str(old.shape), str(new.shape)))
         
            if old.shape != new.shape:
                # we need to recalculate units
                self.make_units()
            
            if name =='_data':
                logger.debug("\nSource : data need replot")
                self.data_need_replot = not self.is_processing

            elif name == '_modeldata':
                logger.debug("\nSource : modeldata need replot")
                self.model_need_replot = not self.is_processing
                    
        else: 
#            logger.debug("Source : %s %s %s"%(name, str(old), str(new)))
            if name == 'unmodified' and not self.is_freq[-1]:
                self.data_need_replot = True 
                
            if name == 'show_model':
                self.model_need_replot = True 
            
            if name == 'show_diff' and self.show_model:
                self.model_need_replot = True 
                
            if name == 'subtract':
                self.data_need_replot = True 
                self.model_need_replot = True
            
            if name == 'forceupdatemodel':
                self.model_need_replot = True
                
            if name == 'is_freq_items':
                self.make_units()
                self.fittable = self.is_freq[-1] 
                if not self.fittable:
                    self.show_model = False
                    
                if new[-1]:
                    logger.debug("reset fit zone and clip setting")
                    #define or redefine the clip limits
                    uc = self.units
                    zone = list(uc[-1].ppm_limits())  
                    # TODO: change this depending on the axis display 
                    self.clip = eval(self.sourcepref.get('processing.clip',
                                                         str(zone)))
                    self.zone = eval(self.sourcepref.get('fitting.zone',
                                                         str(zone)))

    def _is_transposed_changed(self, old, new):
        if old == []:
            return # just initalized do nothing

        if new != old: #True 
            logger.debug("source data transposed?:%s" % str(new))
            # we must exchange the order of the parameters also
            for item in self._attribs:
                val = getattr(self, item) 
                val.reverse()        

    #===========================================================================
    # Object initialisation
    #===========================================================================
    def __init__(self, path=None, **traits):
        """
        Create and setup the Source object
        """

        super(Source, self).__init__(**traits)

        self._is_ok = False  
        self.is_loaded = False

        if path == 'last':
            last = self.applicationpref.get('application.last_source_file')
            if os.path.exists(last):
                self._is_ok = True
                self._directory = last

        elif path is not None:
            if os.path.exists(path):
                self._is_ok = True
                self._directory = path
            else:
                # try by adding path_of_data
                pathofdata = self.applicationpref.get(
                                                    'application.path_of_data')
                path = os.path.join(pathofdata, path)
                if os.path.exists(path):
                    self._is_ok = True
                    self._directory = path
                else:
                    logger.error('cannot find dataset: %s' % path)
        else:    
            # launch the view for selecting the data
            self.edit_traits(view=self._load_view(), kind='modal')

        self._import_bruker_source()   
        logger.debug('Source is_loaded : %s' % str(self.is_loaded))

    #===========================================================================
    # copy
    #===========================================================================
    def copy(self):
        """
        make a copy of this source
        
        """
        return deepcopy(self)

    #===========================================================================
    # _remove_digital_filter
    #===========================================================================
    def _remove_digital_filter(self):
        """Remove the digital filter from Bruker data.
        (adapted from nmrglue.fileio.bruker with some modifications
        as the original routine was not working properly for me. 
        https://github.com/jjhelmus/nmrglue - Copyright (C) 2010 Jonathan Helmus
        License: BSD)
        """

        if self.DECIM is None:
            raise ValueError("par does not contain DECIM parameter")

        decim = self.DECIM[-1]

        if self.DSPFVS  is None:
            raise ValueError("par does not contain DSPFVS parameter")

        dspfvs = self.DSPFVS[-1]

        if self.GRPDLY  is None:
            grpdly = 0
        else:
            grpdly = self.GRPDLY[-1]

        bruker_dsp_table = {
            10: { 
                2    : 44.75,
                3    : 33.5,
                4    : 66.625,
                6    : 59.083333333333333,
                8    : 68.5625,
                12   : 60.375,
                16   : 69.53125,
                24   : 61.020833333333333,
                32   : 70.015625,
                48   : 61.34375,
                64   : 70.2578125,
                96   : 61.505208333333333,
                128  : 70.37890625,
                192  : 61.5859375,
                256  : 70.439453125,
                384  : 61.626302083333333,
                512  : 70.4697265625,
                768  : 61.646484375,
                1024 : 70.48486328125,
                1536 : 61.656575520833333,
                2048 : 70.492431640625,
                },
            11: {
                2    : 46.,
                3    : 36.5,
                4    : 48.,
                6    : 50.166666666666667,
                8    : 53.25,
                12   : 69.5,
                16   : 72.25,
                24   : 70.166666666666667,
                32   : 72.75,
                48   : 70.5,
                64   : 73.,
                96   : 70.666666666666667,
                128  : 72.5,
                192  : 71.333333333333333,
                256  : 72.25,
                384  : 71.666666666666667,
                512  : 72.125,
                768  : 71.833333333333333,
                1024 : 72.0625,
                1536 : 71.916666666666667,
                2048 : 72.03125
                },
            12: {
                2    : 46. ,
                3    : 36.5,
                4    : 48.,
                6    : 50.166666666666667,
                8    : 53.25,
                12   : 69.5,
                16   : 71.625,
                24   : 70.166666666666667,
                32   : 72.125,
                48   : 70.5,
                64   : 72.375,
                96   : 70.666666666666667,
                128  : 72.5,
                192  : 71.333333333333333,
                256  : 72.25,
                384  : 71.666666666666667,
                512  : 72.125,
                768  : 71.833333333333333,
                1024 : 72.0625,
                1536 : 71.916666666666667,
                2048 : 72.03125
                },
            13: {
                2    : 2.75,
                3    : 2.8333333333333333,
                4    : 2.875,
                6    : 2.9166666666666667,
                8    : 2.9375,
                12   : 2.9583333333333333,
                16   : 2.96875,
                24   : 2.9791666666666667,
                32   : 2.984375,
                48   : 2.9895833333333333,
                64   : 2.9921875,
                96   : 2.9947916666666667
                } 
            }

        if grpdly > 0:  # use group delay value if provided (not 0 or -1)
            phase = grpdly

        # determind the phase correction
        else:
            if dspfvs >= 14:  # DSPFVS greater than 14 give no phase correction.
                phase = 0.
            else:   # loop up the phase in the table
                if dspfvs not in bruker_dsp_table:
                    logger.warn("dspfvs %d not in lookup table" % dspfvs) 
                    return False
                if decim not in bruker_dsp_table[dspfvs]:
                    logger.warn("decim %d not in lookup table" % decim)
                    return False

                phase = bruker_dsp_table[dspfvs][decim]

        # and the number of points to remove (skip) and add to the beginning
        skip = int(np.floor(phase + 2.))  # round up two integers
        add = int(max(skip - 6, 0))        # 6 less, or 0

        data = _ft(self._data_orig) # the only masai routine 
                                    # that change data_orig
                                    # we do not want to do this again, anyway
        s = float(data.shape[-1])
        ph = 2.*np.pi * phase * np.arange(s) / s
        data = data * np.exp(1j * ph)
        data = _ift(data)

        # add points at the end of the specta to beginning
        # data[..., :add] = data[..., :add]+data[..., :-(add+1):-1]

        # remove points at end of spectra
        self._data_orig = data[..., :-skip]
        self.skip = skip

        # correct TD
        self.TD[-1] = (self.TD[-1] - skip) / 2

#===============================================================================
# _UnitConversion
#===============================================================================
class _UnitConversion():
    """
    Provides methods to convert between common NMR units
    (adapted from nmrglue.unit_conversion in fileiobase module - 
       https://github.com/jjhelmus/nmrglue - Copyright (C) 2010 Jonathan Helmus
     License: BSD)

    Parameters
    ----------
    size : int
        Number of points in dimension (R|I).
    cplex : bool
        True if dimension is complex, False is real.
    sw : float
        Spectral width in Hz.
    obs : float
        Observation frequency in MHz (SFO1 on Bruker)
    car : float
        Carrier frequency in Hz (SR on bruker)

    """
    def __init__(self, size, cplx, sw, obs, car):
        """
        create and set up a unit_conversion object
        """
        # fundamental units
        self._size = size
        self._cplx = cplx
        self._sw = sw
        self._obs = obs
        self._car = car

        # derived units (these are in ppm)
        self._delta = self._sw / (self._size * self._obs)  
        # CF correction (- --> +) 
        # because spectra are not reversed after fourier transform
        self._first = self._car / self._obs - self._delta * self._size / 2.

    # individual unit conversion functions
    def __percent2pts(self, percent):
        return percent * (self._size - 1) / 100.0

    def __pts2percent(self, pts):
        return pts * 100 / (self._size - 1.0)

    def __hz2pts(self, hz):
        return ((hz / self._obs) - self._first) / self._delta

    def __pts2hz(self, pts):
        return (pts * self._delta + self._first) * self._obs

    def __ppm2pts(self, ppm):
        return (ppm - self._first) / self._delta

    def __pts2ppm(self, pts):
        return (pts * self._delta) + self._first

    # times based units: seconds, ms, and us
    def __sec2pts(self, sec):
        return sec * self._sw

    def __pts2sec(self, pts):
        return pts * 1. / self._sw

    def __ms2pts(self, ms):
        return ms * self._sw / 1.e3

    def __pts2ms(self, pts):
        return pts * 1.e3 / self._sw

    def __us2pts(self, us):
        return us * self._sw / 1.e6

    def __pts2us(self, pts):
        return pts * 1.e6 / self._sw

    # routers
    def __unit2pnt(self, val, units):
        """
        Convert units to points
        """
        units = units.upper()
        if units == "PPM":
            pts = self.__ppm2pts(val)
        elif units == "HZ":
            pts = self.__hz2pts(val)
        elif units == "%" or units == "PERCENT":
            pts = self.__percent2pts(val)
        elif units == "SEC" or units == "SECOND" or units == "S":
            pts = self.__sec2pts(val)
        elif units == "MS":
            pts = self.__ms2pts(val)
        elif units == "US":
            pts = self.__us2pts(val)
        else:
            raise ValueError("invalid unit type")
        #if self._cplx:
        #    return pts+round(pts)
        #else:
        return pts

    def __pnt2unit(self, val, units):
        """
        Convert points to units
        """
        units = units.upper()
        #if self._cplx:
        #    val = val-round(val)
        if units == "PPM":
            k = self.__pts2ppm(val)
        elif units == "HZ":
            k = self.__pts2hz(val)
        elif units == "%" or units == "PERCENT":
            k = self.__pts2percent(val)
        elif units == "SEC" or units == "SECOND" or units == "S":
            k = self.__pts2sec(val)
        elif units == "MS":
            k = self.__pts2ms(val)
        elif units == "US":
            k = self.__pts2us(val)
        else:
            raise ValueError("invalid units")
        return k

    def __str2pnt(self, s):
        """
        Convert string with units to points
        """
        units = s.strip(string.digits + string.whitespace + "." + "-").upper()
        val = float(s.strip(string.ascii_letters + string.whitespace + "%"))
        return self.__unit2pnt(val, units)

    def __convert(self, val, unit=None):
        """
        Convert string or value/unit pair
        """
        if type(val) == str:
            return self.__str2pnt(val)
        else:
            if unit == None:
                raise ValueError("invalid unit type")
            return self.__unit2pnt(val, unit)

    # User functions
    def f(self, val, unit=None):
        """
        Convert string or value/unit pair to float
        """
        return self.__convert(val, unit)

    def i(self, val, unit=None):
        """
        Convert string or value/unit pair to integer
        """
        return int(round(self.__convert(val, unit)))

    def ppm(self, val):
        """
        Convert to ppm
        """
        return self.__pnt2unit(val, "PPM")

    def hz(self, val):
        """
        Convert to Hz
        """
        return self.__pnt2unit(val, "HZ")

    def percent(self, val):
        """
        Convert to percent
        """
        return self.__pnt2unit(val, "PERCENT")

    def seconds(self, val):
        """
        Convert to seconds
        """
        return self.__pnt2unit(val, "SEC")

    def sec(self, val):
        """
        Convert to seconds
        """
        return self.__pnt2unit(val, "SEC")

    def ms(self, val):
        """
        Convert to milliseconds (ms)
        """
        return self.__pnt2unit(val, "MS")

    def us(self, val):
        """
        Convert to microseconds (us)
        """
        return self.__pnt2unit(val, "US")

    def unit(self, val, unit):
        """
        Convert val points to unit
        """
        return self.__pnt2unit(val, unit)

    # limits and scales
    def percent_limits(self):
        """
        Return tuple of left and right edges in percent
        """
        return 0.0, 100.0

    def percent_scale(self):
        """
        Return array of percent values
        """
        return np.linspace(0.0, 100.0, self._size)

    def ppm_limits(self):
        """
        Return tuple of left and right edges in ppm
        """
        return self.ppm(0), self.ppm(self._size - 1)

    def ppm_scale(self):
        """
        Return array of ppm values
        """
        x0, x1 = self.ppm_limits()
        return np.linspace(x0, x1, self._size)

    def hz_limits(self):
        """
        Return tuple of left and right edges in Hz
        """
        return self.hz(0), self.hz(self._size - 1)

    def hz_scale(self):
        """
        Return array of Hz values
        """
        x0, x1 = self.hz_limits()
        return np.linspace(x0, x1, self._size)

    def sec_limits(self):
        """
        Return tuple of left and right edges in seconds
        """
        return self.sec(0), self.sec(self._size - 1)

    def sec_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.sec_limits()
        return np.linspace(x0, x1, self._size)

    def ms_limits(self):
        """
        Return tuple of left and right edges in milliseconds
        """
        return self.ms(0), self.ms(self._size - 1)

    def ms_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.ms_limits()
        return np.linspace(x0, x1, self._size)

    def us_limits(self):
        """
        Return tuple of left and right edges in milliseconds
        """
        return self.us(0), self.us(self._size - 1)

    def us_scale(self):
        """
        Return array of seconds values
        """
        x0, x1 = self.us_limits()
        return np.linspace(x0, x1, self._size)

    __call__ = i    # calling the object x is the same as x.i

#===============================================================================
# ConvertUnits
#===============================================================================
class ConvertUnits(HasTraits, _UnitConversion):
    """ 
    Provides methods to convert between common NMR units
    (subclass of nmrglue.fileio.fileiobase.unit_conversion)

    Examples
    --------

    >>> import pylab as pl
    >>> from masai.api import *
    >>> fig = pl.figure(3)
    >>> fd3 = Source(path='test/1') 
    >>> uc = ConvertUnits(fd3)
    >>> p = pl.plot(uc.ms_scale(), fd3.data.real)
    >>> label = pl.xlabel('ms')

    Parameters
    ----------
    fd : Source Instance
        Contains bruker NMR parameters and data
    dim : int, optional, default=-1
        Dimension number to create unit conversion object for. 
        Default is for last (direct) dimension.

    Attributes
    ----------
    sr : float
        Spectrum reference frequency in Hz
        i.e., SR = SFO1 - SF in Bruker Xwinnmr/Topspin

    """    
    _curdim = Enum(-1, 0, 1, 2)

    sf = Property(Float)

    sr = Property(Float)

    def _get_sr(self):
        return (self.fd.SFO1[self._curdim] - self.sf) * 1.e6

    def _set_sr(self, value):
        self.fd.SF[self._curdim] = self.fd.SFO1[self._curdim] - value * 1.e-6

    def _get_sf(self):
        return self.fd.SF[self._curdim]

    def _set_sf(self, value):
        self.fd.SF[self._curdim] = value

    #===========================================================================
    # __init__
    #===========================================================================
    def __init__(self, fd, dim= -1):
        """
        Setup a Unit conversion object for given dimension.
        """
        self.fd = fd
        self._curdim = dim
        size = float(fd.data.shape[dim])
        ndims = fd.ndims
        is_quad = fd.MC2[-1] > 1

        # check for quadrature in indirect dimentions
        if is_quad and (dim != ndims - 1):
            size = size / 2.
            cplx = True
        else:
            cplx = False

        sw = fd.SW_h[dim]
        sfo1 = self.fd.SFO1[self._curdim]

        _UnitConversion.__init__(self, size, cplx, sw, sfo1, self.sr)




# api
#===============================================================================
# Copyright (C) 2012-2013 Christian Fernandez
#       Laboratoire Catalyse et Spectrochimie, Caen, France.  
#       christian.fernandez@ensicaen.fr
# This software is governed by the CeCILL-B license under French law 
# and abiding by the rules of distribution of free software.  
# You can  use, modify and/ or redistribute the software under 
# the terms of the CeCILL-B license as circulated by CEA, CNRS and INRIA
# at the following URL "http://www.cecill.info".
# See Licence.txt in the main masai source directory
#===============================================================================

"""This module is the main API (Application Programming Interface).

Up to now, only the :ref:`Source <source>` class 
and the :ref:`processing functions <process>` 
are made available to the user.

exemple
-------

    >>> import masai.api
    >>> fd = Source('test/1')  # read the source dataset
    >>> efp(fd)                # performs EM+FT+PK (similar to Bruker commands)


"""

from masai.fileio.source import Source
from masai.options import *
from masai.processing.process import * 




# -*- coding: utf-8 -*-
"""
A fortran domain for sphinx

"""
# Copyright or © or Copr. Actimar (contributor(s) : Stephane Raynaud) (2010)
# 
# raynaud@actimar.fr
# 
# 
# This software is a computer program whose purpose is to provide
# utilities for handling oceanographic and atmospheric data,
# with the ultimate goal of validating the MARS model from IFREMER.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# 

import re

from docutils import nodes
from docutils.parsers.rst import Directive, directives

from sphinx import addnodes
from sphinx.roles import XRefRole
from sphinx.locale import l_, _
from sphinx.domains import Domain, ObjType, Index
from sphinx.directives import ObjectDescription
from sphinx.util.nodes import make_refnode
from sphinx.util.compat import Directive
from sphinx.util.docfields import Field, GroupedField, TypedField, DocFieldTransformer, _is_single_paragraph

# FIXME: surlignage en jaune de la recherche inactive si "/" dans target

# Utilities

def convert_arithm(node, expr, modname=None, nodefmt=nodes.Text):
    """Format an arithmetic expression for a node"""
    ops = re.findall(r'(\W+)',expr)
    nums = re.split(r'\W+', expr)
    if len(nums)!=len(ops): ops.append('')
    for num, op in zip(nums, ops):
        if num:
            if num[0].isalpha():
                refnode = addnodes.pending_xref(
                    '', refdomain='f', reftype='var', reftarget=num,
                        modname=modname)
                refnode += nodefmt(num, num)
                node += refnode
            else:
                node += nodefmt(num, num)
        if op:
            node += nodefmt(op, op)

def parse_shape(shape):
    if not shape: return
    if not shape.startswith('('): shape = '('+shape
    if not shape.endswith(')'): shape += ')'
    return shape
    
def add_shape(node, shape, modname=None, nodefmt=nodes.Text):
    """Format a shape expression for a node"""
    dims = re.split('\s*,\s*', shape.strip('( )'))
    node += nodefmt(' (',' (')
    convert_arithm(node, shape.strip('( )'), modname=modname, nodefmt=nodefmt)
    node += nodefmt(')',')')

#class fortranfield(nodes.Admonition, nodes.TextElement): pass

# Doc fields

    
re_name_shape = re.compile('(\w+)(\(.+\))?')

re_fieldname_match = re.compile(r'(?P<type>\b\w+\b)?\s*(?P<name>\b\w+\b)\s*(?P<shape>\(.*\))?\s*(?P<sattrs>\[.+\])?').match

class FortranField(Field):
    def make_xref(self, rolename, domain, target, innernode=nodes.emphasis,
        modname=None, typename=None):
        if not rolename:
            return innernode(target, target)
        refnode = addnodes.pending_xref('', refdomain=domain, refexplicit=False,
                                        reftype=rolename, reftarget=target,
                                        modname=modname, typename=typename)
        refnode += innernode(target, target)
        return refnode
    

class FortranCallField(FortranField):
    is_grouped = True

    def __init__(self, name, names=(), label=None, rolename=None):
        Field.__init__(self, name, names, label, True, rolename)

    def make_field(self, types, domain, items, **kwargs):
        
        fieldname = nodes.field_name('', self.label)
        #par = Field.make_field(self, types, domain, items[0])
        par = nodes.paragraph()
        for i,item in enumerate(items):
            if i: par += nodes.Text(' ')
            par += item[1]#Field.make_field(self, types, domain, item)
        fieldbody = nodes.field_body('', par)
        return nodes.field('', fieldname, fieldbody)

class FortranCompleteField(FortranField, GroupedField):
    """
    A doc field that is grouped and has type information for the arguments.  It
    always has an argument.  The argument can be linked using the given
    *rolename*, the type using the given *typerolename*.

    Two uses are possible: either parameter and type description are given
    separately, using a field from *names* and one from *typenames*,
    respectively, or both are given using a field from *names*, see the example.

    Example::

       :param foo: description of parameter foo
       :type foo:  SomeClass

       -- or --

       :param SomeClass foo: description of parameter foo
    """
    is_typed = 2

    def __init__(self, name, names=(), typenames=(), label=None,
                 rolename=None, typerolename=None, 
                 shapenames=None, attrnames=None,
                 prefix=None,
                 strong=True,
                 can_collapse=False):
        GroupedField.__init__(self, name, names, label, rolename, can_collapse)
        self.typenames = typenames
        self.typerolename = typerolename
        self.shapenames = shapenames
        self.attrnames = attrnames
        self.prefix = prefix
        if strong:
            self.namefmt = nodes.strong
        else:
            self.namefmt = addnodes.desc_name

    def make_field(self, types, domain, items, shapes=None, attrs=None, 
        modname=None, typename=None):
        def handle_item(fieldarg, content):
            par = nodes.paragraph()
            if self.prefix:
                par += self.namefmt(self.prefix, self.prefix)
                
            par += self.make_xref(self.rolename, domain, fieldarg, self.namefmt, 
                        modname=modname, typename=typename)
            #par += self.namefmt(fieldarg, fieldarg)
            
            fieldtype = types.pop(fieldarg, None)
            fieldshape = shapes and shapes.pop(fieldarg, None)
            fieldattrs = attrs and attrs.pop(fieldarg, None)
            if fieldshape:
                shape = parse_shape(fieldshape[0].astext())
                #par += nodes.Text(' %s'%shape)
                add_shape(par, shape, modname=modname)
            if fieldtype or fieldattrs:
                par += nodes.emphasis(' [',' [' )
            if fieldtype:
                if len(fieldtype) == 1 and isinstance(fieldtype[0], nodes.Text):
                    thistypename = fieldtype[0].astext()
                    #typename = u''.join(n.astext() for n in fieldtype)
                    par += self.make_xref(self.typerolename, domain, thistypename, 
                        modname=modname, typename=typename)
                else:
                    par += fieldtype
            if fieldattrs:
                if fieldtype: par += nodes.emphasis(',',',')
                par += fieldattrs
            if fieldtype or fieldattrs:
                par += nodes.emphasis(']',']')
            if content:
                par += nodes.Text(' :: ')
                par += content
            return par

        if len(items) == 1 and self.can_collapse:
            fieldarg, content = items[0]
            bodynode = handle_item(fieldarg, content)
        else:
            bodynode = self.list_type()
            for fieldarg, content in items:
              bodynode += nodes.list_item('', handle_item(fieldarg, content))
        label = self.label or ''
        fieldname = nodes.field_name('', label)
        fieldbody = nodes.field_body('', bodynode)
        return nodes.field('', fieldname, fieldbody)

    

class FortranDocFieldTransformer(DocFieldTransformer):
    """
    Transforms field lists in "doc field" syntax into better-looking
    equivalents, using the field type definitions given on a domain.
    """

    def __init__(self, directive, modname=None, typename=None):
        self.domain = directive.domain
        if '_doc_field_type_map' not in directive.__class__.__dict__:
            directive.__class__._doc_field_type_map = \
                self.preprocess_fieldtypes(directive.__class__.doc_field_types)
        self.typemap = directive._doc_field_type_map
        self.modname = modname
        self.typename = typename

    def preprocess_fieldtypes(self, types):
        typemap = {}
        for fieldtype in types:
            for name in fieldtype.names:
                typemap[name] = fieldtype, False
            if fieldtype.is_typed:
                for name in fieldtype.typenames:
                    typemap[name] = fieldtype, 'types'
                for name in fieldtype.shapenames:
                    typemap[name] = fieldtype, 'shapes'
                for name in fieldtype.attrnames:
                    typemap[name] = fieldtype, 'attrs'
        return typemap

    def scan_fieldarg(self, fieldname):
        """Extract type, name, shape and attributes from a field name.
        
        :Some possible syntaxes:
        
            - ``p name``
            - ``p type name(shape) [attr1,attr2]``
            - ``p type name``
            - ``p name [attr1, attr2]``
            
        :Returns: ``name, shape, type, list of attributes``.
            if no shape is specified, it is set to ``None``,
        """
        m = re_fieldname_match(fieldname.strip())
        if not m:
            raise ValueError('Wrong field (%s). It must have at least one parameter name and one argument'%fieldname)
        ftype, name, shape, attrs = m.groups()
        attrs = attrs and attrs[1:-1]
        #if attrs:
            #attrs = [a.strip() for a in attrs[1:-1].split(',')]
        #else: 
            #attrs = []
        return name, shape, ftype, attrs

    def transform(self, node):
        """Transform a single field list *node*."""
        typemap = self.typemap
        fmodname = self.modname
        ftypename = self.typename

        entries = []
        groupindices = {}
        types = {}
        shapes = {}
        attrs = {}
        
        # step 1: traverse all fields and collect field types and content
        for field in node:

            fieldname, fieldbody = field
            try:
                # split into field type and argument
                fieldtype, fieldarg = fieldname.astext().split(None, 1)
            except ValueError:
                # maybe an argument-less field type?
                fieldtype, fieldarg = fieldname.astext(), ''
            typedesc, is_typefield = typemap.get(fieldtype, (None, None))

            # sort out unknown fields
            if typedesc is None :#or typedesc.has_arg != bool(fieldarg):
                # either the field name is unknown, or the argument doesn't
                # match the spec; capitalize field name and be done with it
                new_fieldname = fieldtype.capitalize() + ' ' + fieldarg
                fieldname[0] = nodes.Text(new_fieldname)
                entries.append(field)
                continue

            typename = typedesc.name

            # collect the content, trying not to keep unnecessary paragraphs
            if _is_single_paragraph(fieldbody):
                content = fieldbody.children[0].children
            else:
                content = fieldbody.children

            # if the field specifies a type, put it in the types collection
            if is_typefield:
                # filter out only inline nodes; others will result in invalid
                # markup being written out
                content = filter(
                    lambda n: isinstance(n, nodes.Inline) or
                              isinstance(n, nodes.Text),
                    content)
                if content:
                    eval(is_typefield).setdefault(typename, {})[fieldarg] = content
                continue

            # also support syntax like ``:param type name [attrs]:``
            if typedesc.is_typed==2:
                argname, argshape, argtype, argattrs = self.scan_fieldarg(fieldarg)
                if argtype:
                    types.setdefault(typename, {})[argname] = \
                                               [nodes.Text(argtype)]
                if argshape:
                    shapes.setdefault(typename, {})[argname] = \
                                               [nodes.Text(argshape)]
                if argattrs:
                    attrs.setdefault(typename, {})[argname] = \
                                               [nodes.emphasis(argattrs,argattrs)]
                fieldarg = argname
            elif typedesc.is_typed:
                try:
                    argtype, argname = fieldarg.split(None, 1)
                except ValueError:
                    pass
                else:
                    types.setdefault(typename, {})[argname] = \
                                               [nodes.Text(argtype)]
                    fieldarg = argname

            # grouped entries need to be collected in one entry, while others
            # get one entry per field
            if typedesc.is_grouped:
                if typename in groupindices:
                    group = entries[groupindices[typename]]
                else:
                    groupindices[typename] = len(entries)
                    group = [typedesc, []]
                    entries.append(group)
                group[1].append(typedesc.make_entry(fieldarg, content))
            else:
                entries.append([typedesc,
                                typedesc.make_entry(fieldarg, content)])

        # step 2: all entries are collected, construct the new field list
        new_list = nodes.field_list()
        for entry in entries:
            if isinstance(entry, nodes.field):
                # pass-through old field
                new_list += entry
            else:
                fieldtype, content = entry
                fieldtypes = types.get(fieldtype.name, {})
                fieldshapes = shapes.get(fieldtype.name, {})
                fieldattrs = attrs.get(fieldtype.name, {})
                new_list += fieldtype.make_field(fieldtypes, self.domain,
                    content, shapes=fieldshapes, attrs=fieldattrs,
                    modname=fmodname, typename=ftypename)

        node.replace_self(new_list)



# REs for Fortran signatures
f_sep = '/'
f_sig_re = re.compile(
    r'''^ (\w+(?:[^%%%(f_sep)s]%(f_sep)s\w+))? \s*          # type
          (\b(?:subroutine|function))?  \s*             # objtype
          (\b\w+%(f_sep)s)?              # module name
          (\b\w+%%)?              # type name
          (\b\w+)  \s*             # thing name
          (?: \((.*)\))?           # optional: arguments
           $                   # and nothing more
          '''%dict(f_sep=f_sep), re.VERBOSE+re.I)





# Directives

# RE to split at word boundaries
wsplit_re = re.compile(r'(\W+)')
f_type_re = re.compile('^([\w]+).*$')

f_paramlist_re = re.compile(r'([\[\],])')  # split at '[', ']' and ','



#def fortran_rsplit(fullname):
#    items = [item for item in f_separator.findall(fullname)]
#    return ''.join(items[:-2]), items[-1]

def _pseudo_parse_arglist(signode, arglist):
    """"Parse" a list of arguments separated by commas.

    Arguments can have "optional" annotations given by enclosing them in
    brackets.  Currently, this will split at any comma, even if it's inside a
    string literal (e.g. default argument value).
    """
    paramlist = addnodes.desc_parameterlist()
    stack = [paramlist]
    try:
        for argument in arglist.split(','):
            argument = argument.strip()
            ends_open = ends_close = 0
            while argument.startswith('['):
                stack.append(addnodes.desc_optional())
                stack[-2] += stack[-1]
                argument = argument[1:].strip()
            while argument.startswith(']'):
                stack.pop()
                argument = argument[1:].strip()
            while argument.endswith(']'):
                ends_close += 1
                argument = argument[:-1].strip()
            while argument.endswith('['):
                ends_open += 1
                argument = argument[:-1].strip()
            if argument:
                stack[-1] += addnodes.desc_parameter(argument, argument)
            while ends_open:
                stack.append(addnodes.desc_optional())
                stack[-2] += stack[-1]
                ends_open -= 1
            while ends_close:
                stack.pop()
                ends_close -= 1
        if len(stack) != 1:
            raise IndexError
    except IndexError:
        # if there are too few or too many elements on the stack, just give up
        # and treat the whole argument list as one argument, discarding the
        # already partially populated paramlist node
        signode += addnodes.desc_parameterlist()
        signode[-1] += addnodes.desc_parameter(arglist, arglist)
    else:
        signode += paramlist

class FortranObject(ObjectDescription):
    """
    Description of a general Fortran object.
    """
    option_spec = {
        'noindex': directives.flag,
        'module': directives.unchanged,
        'type': directives.unchanged,
        'shape': parse_shape,
        'attrs':directives.unchanged,
    }

    doc_field_types = [
        FortranCompleteField('parameter', label=l_('Parameters'),
                   names=('p', 'param', 'parameter', 'a', 'arg', 'argument'),
                   #rolename='var', 
                   typerolename='type', 
                   typenames=('paramtype', 'type', 'ptype'),
                   shapenames=('shape', 'pshape'),
                   attrnames=('attrs', 'pattrs', 'attr'),
                   can_collapse=True),
        FortranCompleteField('optional', label=l_('Options'),
                   names=('o', 'optional', 'opt', 'keyword', 'option'),
                   #rolename='var', 
                   typerolename='type', 
                   typenames=('optparamtype', 'otype'),
                   shapenames=('oshape',),
                   attrnames=('oattrs', 'oattr'),
                   can_collapse=True),
        FortranCompleteField('typefield', label=l_('Type fields'),
                   names=('f', 'field', 'typef', 'typefield'),
                   #rolename='typef', 
                   typerolename='type', 
                   typenames=('fieldtype', 'ftype'),
                   shapenames=('fshape',),
                   attrnames=('fattrs', 'fattr'),
                   prefix='% ',
                   strong=False,
                   can_collapse=False),
        FortranCompleteField('return', label=l_('Return'),
                   names=('r', 'return', 'returns'),
                   typerolename='type', 
                   typenames=('returntype', 'rtype'),
                   shapenames=('rshape',),
                   attrnames=('rattrs', 'rattr'),
                   can_collapse=True),
        FortranCallField('calledfrom', label=l_('Called from'),
            names=('calledfrom', 'from')),
        FortranCallField('callto', label=l_('Call to'),
            names=('callto', 'to')),
    ]

    # These Fortran types aren't described anywhere, so don't try to create
    # a cross-reference to them
    stopwords = set(('float', 'integer', 'character', 'double', 'long'))

    _parens = ''
    
#    def _parse_type(self, node, ftype):
#        m = f_type_re.match(ftype)
#        tnode = nodes.Text(ftype, ftype)
#        modname = self.options.get(
#            'module', self.env.temp_data.get('f:module'))
#        if m :
#            ftype = m.groups(0)
#            if ftype not in self.stopwords:
#                pnode = addnodes.pending_xref(
#                    '', refdomain='f', reftype='type', reftarget=ftype,
#                    modname=modname)
#                pnode += tnode
#                node += pnode
#            else:
#                node += tnode
#        else:
#            node += tnode

    def get_signature_prefix(self, sig):
        """
        May return a prefix to put before the object name in the signature.
        """
        return ''

    def needs_arglist(self):
        """
        May return true if an empty argument list is to be generated even if
        the document contains none.
        """
        return False

    def handle_signature(self, sig, signode):
        """
        Transform a Fortran signature into RST nodes.
        Returns (fully qualified name of the thing, classname if any).

        If inside a class, the current class name is handled intelligently:
        * it is stripped from the displayed name if present
        * it is added to the full name (return value) if not present
        """
        m = f_sig_re.match(sig)
        if m is None:
            raise ValueError
        ftype, objtype, modname, typename, name, arglist = m.groups()
        if not typename: typename = ""
            
        # determine module, type, shape and attributes
        modname = (modname and modname[:-1]) or self.options.get(
            'module', self.env.temp_data.get('f:module'))
        if typename:
            name =  typename[:-1]
        attrs = self.options.get('attrs')
        shape = parse_shape(self.options.get('shape'))
        ftype = ftype or self.options.get('type')
        if self.objtype=='typefield' and not typename:
            raise ValueError
        
        #if typename: name = typename+'%'+name
        
        #fullname = name
        #if modname:
        if self.objtype=='program':
            fullname = name
        else:
            fullname = (modname or '_') + f_sep + name
 
        signode['module'] = modname
        signode['type'] = typename
        signode['fullname'] = fullname

        # Add "function" or "subroutine" tag
        sig_prefix = self.get_signature_prefix(sig)
        if objtype or sig_prefix:
            objtype = objtype or sig_prefix
            signode += addnodes.desc_annotation(objtype+' ', objtype+' ')

        # Add module
        if self.env.config.add_module_names and modname and self.objtype!='typefield':
            nodetext = modname + f_sep
            signode += addnodes.desc_addname(nodetext, nodetext)

        # Add name
        signode += addnodes.desc_name(name, name)
        
        # In the parenthesis
        if self.needs_arglist(): # call for functions and subroutines
            if arglist: # Calling arguments
                _pseudo_parse_arglist(signode, arglist)               
            elif self.needs_arglist():  # for callables, add an empty parameter list       
                signode += addnodes.desc_parameterlist()
        elif arglist and not shape: # Declare shape instead of arguments (variables)
            shape = arglist
            
        # Add remaining
        self.add_shape_and_attrs(signode, modname, ftype, shape, attrs)        
            
        return fullname, ftype

        
        
    def add_shape_and_attrs(self, signode, modname, ftype, shape, attrs):
        # add shape 
        if shape:
            add_shape(signode, shape, modname=modname)
            #signode += nodes.Text(' '+shape)
        # add type ('float', 'interger', etc)
        if ftype or attrs:
            signode += nodes.emphasis(' [', ' [')
        if ftype:
            refnode = addnodes.pending_xref(
                '', refdomain='f', reftype='type', reftarget=ftype,
                modname=modname,)
            refnode += nodes.emphasis(ftype, ftype)
            signode += refnode
            #tnode = addnodes.desc_type(ftype, ftype)
            #tnode +=
            #signode += addnodes.desc_type(ftype, ftype)
            #signode += 
    #        signode += addnodes.desc_type('', '')
    #        self._parse_type(signode[-1], ftype)
        if attrs:
            if ftype: signode += nodes.emphasis(',',',')
            for iatt,att in enumerate(re.split('\s*,\s*', attrs)):
                if iatt:
                    signode += nodes.emphasis(',',',')
                if att.startswith('parameter'):
                    value = att.split('=')[1]
                    signode += nodes.emphasis('parameter=', 'parameter=')
                    convert_arithm(signode, value, modname=modname)
                else:
                    signode += nodes.emphasis(att,att)
            #signode += nodes.emphasis(attrs, attrs)

        if ftype or attrs:
            signode += nodes.emphasis(']', ']')
    def add_target_and_index(self, name, sig, signode):
        #modname = self.options.get(
            #'module', self.env.temp_data.get('f:module'))
        modname = signode.get(
            'module', self.env.temp_data.get('f:module'))
#        fullname = (modname and modname + '/' or '') + name[0]
        fullname = 'f' + f_sep + name[0]

        # note target
        if fullname not in self.state.document.ids:
            signode['names'].append(fullname)
            signode['ids'].append(fullname)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)
            objects = self.env.domaindata['f']['objects']
            if fullname in objects:
                self.env.warn(
                    self.env.docname,
                    'duplicate object description of %s, ' % fullname +
                    'other instance in ' +
                    self.env.doc2path(objects[fullname][0]),
                    self.lineno)
            objects[fullname] = (self.env.docname, self.objtype)
        indextext = self.get_index_text(modname, fullname)
        if indextext:
            self.indexnode['entries'].append(('single', indextext,
                                              fullname, fullname))

    def before_content(self):
        # needed for automatic qualification of fields (reset in subclasses)
        self.typename_set = False

    def after_content(self):
        if self.typename_set:
            self.env.temp_data['f:type'] = None
            
    def get_index_text(self, modname, name):
        add_modules = self.env.config.add_module_names
        if name.startswith('f'+f_sep): name = name[2:]
        mn = modname or '_'
        sobj = ''
        if name.startswith(mn+f_sep):
            name = name[len(mn)+1:]
        if self.objtype=='type':
            sobj = _('fortran type')
        if self.objtype=='typefield':
            sobj = _('fortran type field')
        elif self.objtype=='variable':
            sobj = _('fortran variable')
        elif self.objtype=='subroutine':
            sobj = _('fortran subroutine')
        elif self.objtype=='function':
            sobj = _('fortran function')
        elif self.objtype=='module':
            sobj = _('fortran module')
            modname = ''
        elif self.objtype=='program':
            sobj = _('fortran program')
            modname = ''
        sinmodule = (_(' in module %s')%modname) if modname and add_modules else ''
        return '%s%s (%s%s)'%(name, self._parens, sobj, sinmodule)

class FortranSpecial:
    def get_signature_prefix(self, sig):
        """
        May return a prefix to put before the object name in the signature.
        """
        return self.objtype+' '
    
class WithFortranDocFieldTransformer:
    def run(self):
        """Same as :meth:`sphinx.directives.ObjectDescription` 
        but using :class:`FortranDocFieldTransformer`"""
        if ':' in self.name:
            self.domain, self.objtype = self.name.split(':', 1)
        else:
            self.domain, self.objtype = '', self.name
        self.env = self.state.document.settings.env
        self.indexnode = addnodes.index(entries=[])

        node = addnodes.desc()
        node.document = self.state.document
        node['domain'] = self.domain
        # 'desctype' is a backwards compatible attribute
        node['objtype'] = node['desctype'] = self.objtype
        node['noindex'] = noindex = ('noindex' in self.options)

        self.names = []
        signatures = self.get_signatures()
        for i, sig in enumerate(signatures):
            # add a signature node for each signature in the current unit
            # and add a reference target for it
            signode = addnodes.desc_signature(sig, '')
            signode['first'] = False
            node.append(signode)
            try:
                # name can also be a tuple, e.g. (classname, objname);
                # this is strictly domain-specific (i.e. no assumptions may
                # be made in this base class)
                name = self.handle_signature(sig, signode)
            except ValueError:
                # signature parsing failed
                signode.clear()
                signode += addnodes.desc_name(sig, sig)
                continue  # we don't want an index entry here
            if not noindex and name not in self.names:
                # only add target and index entry if this is the first
                # description of the object with this name in this desc block
                self.names.append(name)
                self.add_target_and_index(name, sig, signode)

        modname = signode.get('module')
        typename = signode.get('type')
        contentnode = addnodes.desc_content()
        node.append(contentnode)
        if self.names:
            # needed for association of version{added,changed} directives
            self.env.temp_data['object'] = self.names[0]
        self.before_content()
        self.state.nested_parse(self.content, self.content_offset, contentnode)
        FortranDocFieldTransformer(self, modname=modname, typename=typename).transform_all(contentnode)
        self.env.temp_data['object'] = None
        self.after_content()
        return [self.indexnode, node]



class FortranType(FortranSpecial, WithFortranDocFieldTransformer, FortranObject):
    def before_content(self):
        FortranObject.before_content(self)
        if self.names:
            self.env.temp_data['f:type'] = self.names[0][0].split(f_sep)[-1]
            self.typename_set = True

class FortranTypeField(FortranObject):
    #def handle_signature(self, sig, signode):
        #"""
        #Transform a Fortran signature into RST nodes.
        #Returns (fully qualified name of the thing, classname if any).

        #If inside a class, the current class name is handled intelligently:
        #* it is stripped from the displayed name if present
        #* it is added to the full name (return value) if not present
        #"""
        #m = f_sig_re.match(sig)
        #if m is None:
            #raise ValueError
        #ftype, objtype, modname, typename, name, arglist = m.groups()
        #print 'handle_signature', ftype, objtype, modname, typename, name, arglist 
        #if not typename: typename = ""
            
        ## determine module and type
        #modname = (modname and modname[:-1]) or self.options.get(
            #'module', self.env.temp_data.get('f:module'))
        #typename = (typename and typename[:-1]) or self.options.get(
            #'type', self.env.temp_data.get('f:type'))
        ##print ' mod type', modname, typename
        ##print self.objtype
        #if self.objtype=='typefield' and not typename:
            #raise ValueError
        
        #if typename: name = typename+'%'+name
        
        #fullname = name
        #if modname:
            #fullname = modname + f_sep + name
 
        #signode['module'] = modname
        #signode['type'] = typename
        #signode['fullname'] = fullname
        
        ## Fill node
        #signode += addnodes.desc_name(name, name)
        #shape = self.options.get('shape')
        #if shape: signode += nodes.Text(shape, shape)
        #ftype = self.options.get('type', ftype)
        #attr= self.options.get('attr')
        #if ftype or attr:
            #signode += nodes.Text(' :: ', ' :: ')
            #if ftype: signode += nodes.emphasis('', ftype)
            #if attr: signode += nodes.literal('', '['+attr+']')
        #if self.content:
            #signode += nodes.Text(': ', ': ')
            #argnodes, msgs = self.state.inline_text(' '.join(self.content), self.lineno)
            #signode += argnodes
            #signode += msgs

        #return fullname, ftype
        
    def before_content(self):
        FortranObject.before_content(self)
        lastname = self.names and self.names[-1][1]
        if lastname and not self.env.temp_data.get('f:type'):
            self.env.temp_data['f:type'] = lastname.split(f_sep)[-1]
            self.typename_set = True
    
class FortranProgram(FortranSpecial, WithFortranDocFieldTransformer, FortranObject):
    pass
    
class FortranWithSig(FortranSpecial, WithFortranDocFieldTransformer, FortranObject):
    """
    Description of a function of subroutine
    """
    _parens = '()'
    def needs_arglist(self):
        return True
    
    def get_signature_prefix(self, sig):
        """
        May return a prefix to put before the object name in the signature.
        """
        return self.objtype+' '
        
class FortranField(Directive):
    """
    Directive to describe a change/addition/deprecation in a specific version.
    """

    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        'type': directives.unchanged, 
        'shape': parse_shape, 
        'attrs': directives.unchanged, 
        }

    def run(self):
        from docutils import nodes
        node = nodes.paragraph()
        node += addnodes.desc_name(self.arguments[0], self.arguments[0])
        shape = self.options.get('shape')
        if shape :
            #node += nodes.Text(shape, shape)
            add_shape(node, shape)
        type = self.options.get('type')
        attrs= self.options.get('attrs')
        if type or attrs:
            node += nodes.Text(' :: ', ' :: ')
            if type: node += nodes.emphasis('', type)
            if attr: node += nodes.literal('', '['+attr+']')
        if self.content:
            node += nodes.Text(': ', ': ')
            argnodes, msgs = self.state.inline_text(' '.join(self.content), self.lineno)
            node += argnodes
            node += msgs
        ret = [node]
#        env = self.state.document.settings.env
#        env.note_versionchange(node['type'], node['version'], node, self.lineno)
        return ret




class FortranModule(Directive):
    """
    Directive to mark description of a new module.
    """

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        'platform': lambda x: x,
        'synopsis': lambda x: x,
        'noindex': directives.flag,
        'deprecated': directives.flag,
    }

    def run(self):
        env = self.state.document.settings.env
        modname = self.arguments[0].strip()
        noindex = 'noindex' in self.options
        env.temp_data['f:module'] = modname
        env.domaindata['f']['modules'][modname] = \
            (env.docname, self.options.get('synopsis', ''),
             self.options.get('platform', ''), 'deprecated' in self.options)
        env.domaindata['f']['objects']['f'+f_sep+modname] = (env.docname, 'module')
        #targetnode = nodes.target('', '', ids=['module-' + modname], ismod=True)
        targetnode = nodes.target('', '', ids=['f'+f_sep+modname], ismod=True)
        self.state.document.note_explicit_target(targetnode)
        ret = [targetnode]
        # XXX this behavior of the module directive is a mess...
        if 'platform' in self.options:
            platform = self.options['platform']
            node = nodes.paragraph()
            node += nodes.emphasis('', _('Platforms: '))
            node += nodes.Text(platform, platform)
            ret.append(node)
        # the synopsis isn't printed; in fact, it is only used in the
        # modindex currently
        if not noindex:
            indextext = _('%s (module)') % modname
            #inode = addnodes.index(entries=[('single', indextext,
                                             #'module-' + modname, modname)])
            inode = addnodes.index(entries=[('single', indextext,
                                             'f' + f_sep + modname, modname)])
            ret.append(inode)
        return ret


class FortranCurrentModule(Directive):
    """
    This directive is just to tell Sphinx that we're documenting
    stuff in module foo, but links to module foo won't lead here.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = {}

    def run(self):
        env = self.state.document.settings.env
        modname = self.arguments and (self.arguments[0] or self.arguments[0].strip()) or None
        if modname:
            env.temp_data['f:module'] = None
        else:
            env.temp_data['f:module'] = modname
        return []


class FortranXRefRole(XRefRole):
    def process_link(self, env, refnode, has_explicit_title, title, target):
        refnode['f:module'] = env.temp_data.get('f:module')
        refnode['f:type'] =  env.temp_data.get('f:type')
        if not has_explicit_title:
            title = title.lstrip('.')   # only has a meaning for the target
            target = target.lstrip('~') # only has a meaning for the title
            # if the first character is a tilde, don't display the module/class
            # parts of the contents
            if title[0:1] == '~':
                title = title[1:].split(f_sep)[-1]
            #elif '%' not in title and refnode['f:type']:
                #title = '%s%%%s'%(refnode['f:type'], title)
        # if the first character is a sep, search more specific namespaces first
        # else search builtins first
        if target.startswith(f_sep):
            target = target[1:]
            refnode['refspecific'] = True
        return title, target
        


class FortranModuleIndex(Index):
    """
    Index subclass to provide the Fortran module index.
    """

    name = 'modindex'
    localname = l_('Fortran Module Index')
    shortname = l_('fortran modules')

    def generate(self, docnames=None):
        content = {}
        # list of prefixes to ignore
        ignores = self.domain.env.config['modindex_common_prefix']
        ignores = sorted(ignores, key=len, reverse=True)
        # list of all modules, sorted by module name
        modules = sorted(self.domain.data['modules'].iteritems(),
             key=lambda x: x[0].lower())
        # sort out collapsable modules
        prev_modname = ''
        num_toplevels = 0
        for modname, (docname, synopsis, platforms, deprecated) in modules:
            if docnames and docname not in docnames:
                continue

            for ignore in ignores:
                if modname.startswith(ignore):
                    modname = modname[len(ignore):]
                    stripped = ignore
                    break
            else:
                stripped = ''

            # we stripped the whole module name?
            if not modname:
                modname, stripped = stripped, ''

            entries = content.setdefault(modname[0].lower(), [])

            package = modname.split(f_sep)[0]
            if package != modname:
                # it's a submodule
                if prev_modname == package:
                    # first submodule - make parent a group head
                    entries[-1][1] = 1
                elif not prev_modname.startswith(package):
                    # submodule without parent in list, add dummy entry
                    entries.append([stripped + package, 1, '', '', '', '', ''])
                subtype = 2
            else:
                num_toplevels += 1
                subtype = 0

            qualifier = deprecated and _('Deprecated') or ''
            #entries.append([stripped + modname, subtype, docname,
                            #'module-' + stripped + modname, platforms,
                            #qualifier, synopsis])
            entries.append([stripped + modname, subtype, docname,
                            'f' + f_sep + stripped + modname, platforms,
                            qualifier, synopsis or ''])
            prev_modname = modname

        # apply heuristics when to collapse modindex at page load:
        # only collapse if number of toplevel modules is larger than
        # number of submodules
        collapse = len(modules) - num_toplevels < num_toplevels

        # sort by first letter
        content = sorted(content.iteritems())

        return content, collapse


class FortranDomain(Domain):
    """Fortran language domain."""
    name = 'f'
    label = 'Fortran'
    object_types = {
        'program':      ObjType(l_('program'), 'prog'),
        'type':         ObjType(l_('type'), 'type'),
        'variable':     ObjType(l_('variable'), 'var'),
        'function':     ObjType(l_('function'), 'func'),
        'subroutine':   ObjType(l_('subroutine'), 'func', 'subr'),
        'module':       ObjType(l_('module'), 'mod'),
    }

    directives = {
        'program': FortranProgram,
        'type':           FortranType,
        'variable':          FortranObject, 
        'function':        FortranWithSig,
        'subroutine':          FortranWithSig,
        'module':          FortranModule,
        'currentmodule':   FortranCurrentModule,
    }

    roles = {
        'prog': FortranXRefRole(),
        'type': FortranXRefRole(),
        'var':FortranXRefRole(),
        'func':  FortranXRefRole(fix_parens=True),
        'subr':  FortranXRefRole(fix_parens=True),
        'mod':   FortranXRefRole(),
    }
    initial_data = {
        'objects': {},  # fullname -> docname, objtype
        'modules': {},  # modname -> docname, synopsis, platform, deprecated
    }
    indices = [
        FortranModuleIndex,
    ]

    def clear_doc(self, docname):
        for fullname, (fn, _) in self.data['objects'].items():
            if fn == docname:
                del self.data['objects'][fullname]
        for modname, (fn, _, _, _) in self.data['modules'].items():
            if fn == docname:
                del self.data['modules'][modname]

    def find_obj(self, env, modname, name, role, searchorder=0):
        """
        Find a Fortran object for "name", perhaps using the given module and/or
        typename.
        
        :Params:
        
            - **searchorder**, optional: Start using relative search
        """
        # skip parens
        if name.endswith('()'):
            name = name[:-2]

        if not name:
            return None, None
        if f_sep in name:
            modname, name = name.split(f_sep)
        #modname = modname or '_'
        if '%' in name:
            name, tmp = name.split('%')
            
        objects = self.data['objects']
        newname = None
        matches = []
        objtypes = self.objtypes_for_role(role)
        if searchorder == 1: # :role:`/toto`
            if role in ['mod', 'prog']:
                if 'f'+f_sep + name not in objects: # exact match
                    return []
                newname = 'f'+f_sep + name
            elif modname and 'f'+f_sep + modname + f_sep + name in objects and \
               objects['f'+f_sep + modname + f_sep + name][1] in objtypes:
                newname = 'f'+f_sep + modname + f_sep + name
            elif 'f'+f_sep + '_' + f_sep + name in objects and \
               objects['f'+f_sep + '_' + f_sep + name][1] in objtypes:
                newname = 'f'+f_sep + '_' + f_sep + name
            elif 'f'+f_sep + name in objects and \
               objects['f'+f_sep  + name][1] in objtypes:
                newname = 'f'+f_sep + name
            elif  name in objects and \
               objects[name][1] in objtypes:
                newname = name
            
        else: # :role:`toto`
            # NOTE: searching for exact match, object type is not considered
            if 'f'+f_sep + name in objects:
                newname = 'f'+f_sep + name
            elif role in ['mod', 'prog']:
                # only exact matches allowed for modules
                return []
            elif 'f'+f_sep + '_' + f_sep + name in objects:
                 newname = 'f' + f_sep + '_' + f_sep +name
            elif modname and 'f'+f_sep + modname + f_sep + name in objects:
                newname = 'f' + f_sep + modname + f_sep +name
                
        # Last chance: fuzzy search
        if newname is None:
            matches = [(oname, objects[oname]) for oname in objects
               if oname.endswith(f_sep+name)
                   and objects[oname][1] in objtypes]
        else:
            matches.append((newname, objects[newname]))
        return matches




    def resolve_xref(self, env, fromdocname, builder,
                     type, target, node, contnode):
        modname = node.get('f:module', node.get('modname'))
        typename = node.get('f:type', node.get('typename'))
        searchorder = node.hasattr('refspecific') and 1 or 0
        matches = self.find_obj(env, modname, target, type, searchorder)
        if not matches:
            return None
        elif len(matches) > 1:
            env.warn(fromdocname,
                     'more than one target found for cross-reference '
                     '%r: %s' % (target,
                                 ', '.join(match[0] for match in matches)),
                     node.line)
        name, obj = matches[0]

        if obj[1] == 'module':
            # get additional info for modules
            docname, synopsis, platform, deprecated = self.data['modules'][name[1+len(f_sep):]]
            assert docname == obj[0]
            title = name
            if synopsis:
                title += ': ' + synopsis
            if deprecated:
                title += _(' (deprecated)')
            #return make_refnode(builder, fromdocname, docname,
                                #'module-' + name, contnode, title)
            return make_refnode(builder, fromdocname, docname,
                                name, contnode, title)
        else:
            return make_refnode(builder, fromdocname, obj[0], name,
                                contnode, name)



    def get_objects(self):
        for modname, info in self.data['modules'].iteritems():
            yield (modname, modname, 'module', info[0], 'module-' + modname, 0)
        for refname, (docname, type) in self.data['objects'].iteritems():
            yield (refname, refname, type, docname, refname, 1)


def setup(app):
    app.add_domain(FortranDomain)
